from wizzi_utils.google import google_tools as got
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.misc.test import test_misc_tools as mtt
from wizzi_utils.socket import socket_tools as st
import os


def build_gh() -> got.google_handler:
    gh = got.google_handler(
        yaml_path='./GDrive2/settings.yaml',
        tabs=1,
        dir_color='reverse',
        file_color='light_blue',
        title_color=['underlined', 'light_yellow']

    )
    print(gh)
    return gh


def create_test_dir(gh: got.google_handler) -> str:
    print('Creating dir for testing:')
    new_dir_path = 'root/{}'.format(os.path.basename(mtt.PLAY_GROUND))  # new dir
    gh.create_dir(dst_full_path_on_drive=new_dir_path)
    return new_dir_path


def delete_test_dir(gh: got.google_handler, test_f: str) -> None:
    # delete dir for tests
    print('Deleting dir for testing:')
    gh.delete_empty_dir(full_path_on_drive=test_f)
    # gh.delete_dir_with_files(full_path_on_drive=new_dir_path)
    return


def list_all_files_test(gh: got.google_handler = None, test_f: str = None):
    mt.get_function_name(ack=True, tabs=0)
    local_test = False
    if gh is None:
        gh = build_gh()
        test_f = create_test_dir(gh)
        local_test = True
    gh.list_all_files()  # root
    gh.list_all_files(starting_dir=test_f)
    # gh.list_all_files(starting_dir='root/fake_dir')  # will fail
    if local_test:
        delete_test_dir(gh, test_f)
    return


def upload_delete_image_test(gh: got.google_handler = None, test_f: str = None):
    mt.get_function_name(ack=True, tabs=0)
    local_test = False
    if gh is None:
        gh = build_gh()
        test_f = create_test_dir(gh)
        local_test = True
    src_pth = '{}/{}.jpg'.format(mtt.IMAGES_INPUTS, mtt.SO_LOGO)
    if not os.path.exists(src_pth):
        mt.create_dir(os.path.dirname(src_pth))
        url = mtt.IMAGES_D[mtt.SO_LOGO]
        st.download_file(url, dst_path=src_pth)
    dst_full_path_on_drive = '{}/{}.jgp'.format(test_f, mtt.SO_LOGO)  # new file
    gh.upload_file(dst_full_path_on_drive=dst_full_path_on_drive, local_file_path=src_pth)
    gh.delete_file(full_path_on_drive=dst_full_path_on_drive)

    if local_test:
        delete_test_dir(gh, test_f)
    return


def upload_download_delete_file_test(gh: got.google_handler = None, test_f: str = None):
    mt.get_function_name(ack=True, tabs=0)
    local_test = False
    if gh is None:
        gh = build_gh()
        test_f = create_test_dir(gh)
        local_test = True
    src_path = '{}/{}'.format(mtt.TEMP_FOLDER1, mtt.DEMO_FILE)
    mtt.create_demo_file(src_path)
    dst_full_path_on_drive = '{}/{}'.format(test_f, mtt.DEMO_FILE)  # new file

    gh.upload_file(dst_full_path_on_drive=dst_full_path_on_drive, local_file_path=src_path)
    # gh.upload_file(dst_full_path_on_drive=dst_full_path_on_drive, local_file_path=mtt.TEMP_TXT_PATH)  # will fail
    mt.delete_file(src_path)

    gh.download_file(full_path_on_drive=dst_full_path_on_drive, local_save_path=src_path)
    mt.delete_file(src_path)

    gh.delete_file(full_path_on_drive=dst_full_path_on_drive)
    # gh.delete_file(full_path_on_drive=dst_full_path_on_drive)  # will fail
    if local_test:
        delete_test_dir(gh, test_f)
    return


def upload_read_download_delete_test(gh: got.google_handler = None, test_f: str = None):
    mt.get_function_name(ack=True, tabs=0)
    local_test = False
    if gh is None:
        gh = build_gh()
        test_f = create_test_dir(gh)
        local_test = True
    dst_full_path_on_drive = '{}/{}'.format(test_f, mtt.DEMO_FILE)  # new file

    content_txt = '\t\thello world'
    content_txt += '\n\t\tnew server ip is 1.1.1.1:2000'
    gh.upload_content_to_new_file(dst_full_path_on_drive=dst_full_path_on_drive, content=content_txt)
    # try_again to see it fails due to duplicated name
    # gh.upload_content_to_new_file(dst_full_path_on_drive=dst_full_path_on_drive, content=content_txt)  # will fail

    # _ = gh.read_file(full_path_on_drive='{}/fake_file.txt'.format(test_f))  # fake file  # will fail
    test_txt_str = gh.read_file(full_path_on_drive=dst_full_path_on_drive)
    print('\tcontent:')
    print(test_txt_str)
    dst_path = '{}/{}'.format(mtt.TEMP_FOLDER1, mtt.DEMO_FILE)
    gh.download_file(full_path_on_drive=dst_full_path_on_drive, local_save_path=dst_path)
    mt.delete_dir_with_files(mtt.TEMP_FOLDER1)

    gh.delete_file(full_path_on_drive=dst_full_path_on_drive)
    if local_test:
        delete_test_dir(gh, test_f)
    return


def create_and_delete_empty_dir_test(gh: got.google_handler = None, test_f: str = None):
    mt.get_function_name(ack=True, tabs=0)
    local_test = False
    if gh is None:
        gh = build_gh()
        test_f = create_test_dir(gh)
        local_test = True
    dst_full_path_on_drive = '{}/{}'.format(test_f, os.path.basename(mtt.TEMP_FOLDER1))  # new dir
    gh.create_dir(dst_full_path_on_drive=dst_full_path_on_drive)
    gh.delete_empty_dir(full_path_on_drive=dst_full_path_on_drive)
    if local_test:
        delete_test_dir(gh, test_f)
    return


def create_and_delete_dir_with_files_test(gh: got.google_handler = None, test_f: str = None):
    mt.get_function_name(ack=True, tabs=0)
    local_test = False
    if gh is None:
        gh = build_gh()
        test_f = create_test_dir(gh)
        local_test = True
    dst_full_path_on_drive = '{}/{}'.format(test_f, os.path.basename(mtt.TEMP_FOLDER1))  # new dir
    gh.create_dir(dst_full_path_on_drive=dst_full_path_on_drive)
    new_file_dst = '{}/{}'.format(dst_full_path_on_drive, mtt.DEMO_FILE)
    gh.upload_content_to_new_file(dst_full_path_on_drive=new_file_dst, content='\t\thello world')
    gh.list_all_files(starting_dir=dst_full_path_on_drive)
    gh.delete_dir_with_files(full_path_on_drive=dst_full_path_on_drive)
    if local_test:
        delete_test_dir(gh, test_f)
    return


def download_dir_test(gh: got.google_handler = None, test_f: str = None):
    mt.get_function_name(ack=True, tabs=0)
    local_test = False
    if gh is None:
        gh = build_gh()
        test_f = create_test_dir(gh)
        local_test = True
    print('\tuploading some files for the test:')
    new_dir = '{}/{}'.format(test_f, os.path.basename(mtt.TEMP_FOLDER1))  # new dir
    gh.create_dir(dst_full_path_on_drive=new_dir, tabs=2)
    new_file_dst = '{}/{}'.format(new_dir, mtt.DEMO_FILE)
    new_file_dst2 = '{}/v2_{}'.format(new_dir, mtt.DEMO_FILE)
    gh.upload_content_to_new_file(dst_full_path_on_drive=new_file_dst, content='\t\thello world', tabs=3)
    gh.upload_content_to_new_file(dst_full_path_on_drive=new_file_dst2, content='\t\thello world', tabs=3)
    inner_dir = '{}/{}'.format(new_dir, os.path.basename(mtt.TEMP_FOLDER1))  # new dir 2
    gh.create_dir(dst_full_path_on_drive=inner_dir, tabs=3)
    new_file_dst3 = '{}/{}'.format(inner_dir, mtt.DEMO_FILE)
    gh.upload_content_to_new_file(dst_full_path_on_drive=new_file_dst3, content='\t\thello world', tabs=4)
    gh.list_all_files(starting_dir=new_dir)

    # existing dir will fail
    # mt.create_dir(dir_path=mtt.TEMP_FOLDER_NAME)
    # gh.download_dir(full_path_on_drive=new_dir, local_dir_name=mtt.TEMP_FOLDER_NAME)
    # mt.delete_empty_dir(dir_path=mtt.TEMP_FOLDER_NAME)

    gh.download_dir(full_path_on_drive=new_dir, local_dir_name=mtt.TEMP_FOLDER1)
    gh.delete_dir_with_files(full_path_on_drive=new_dir)
    mt.delete_dir_with_files(mtt.TEMP_FOLDER1)
    if local_test:
        delete_test_dir(gh, test_f)
    return


def upload_dir_test(gh: got.google_handler = None, test_f: str = None):
    mt.get_function_name(ack=True, tabs=0)
    local_test = False
    if gh is None:
        gh = build_gh()
        test_f = create_test_dir(gh)
        local_test = True
    print('\tcreating some local files for the test:')
    new_dir = mtt.TEMP_FOLDER1
    mtt.create_demo_file('{}/{}'.format(new_dir, mtt.DEMO_FILE), tabs=2)
    mtt.create_demo_file('{}/v2_{}'.format(new_dir, mtt.DEMO_FILE), tabs=2)
    inner_dir = '{}/{}'.format(new_dir, os.path.basename(mtt.TEMP_FOLDER1))  # new dir 2
    mtt.create_demo_file('{}/{}'.format(inner_dir, mtt.DEMO_FILE), tabs=3)

    # gh.upload_dir(dst_full_path_on_drive=test_f, local_dir_name=mtt.TEMP_FAKE_PATH)  # will fail
    gh.upload_dir(dst_full_path_on_drive='{}/{}'.format(test_f, os.path.basename(new_dir)), local_dir_name=new_dir,
                  tabs=1)
    gh.delete_dir_with_files(full_path_on_drive='{}/{}'.format(test_f, os.path.basename(new_dir)))
    mt.delete_dir_with_files(dir_path=new_dir)
    if local_test:
        delete_test_dir(gh, test_f)
    return


def rename_file_test(gh: got.google_handler = None, test_f: str = None):
    mt.get_function_name(ack=True, tabs=0)
    local_test = False
    if gh is None:
        gh = build_gh()
        test_f = create_test_dir(gh)
        local_test = True
    new_file_dst = '{}/{}'.format(test_f, mtt.DEMO_FILE)
    gh.upload_content_to_new_file(dst_full_path_on_drive=new_file_dst, content='\t\thello world')
    gh.list_all_files(starting_dir=test_f)
    new_name = 'v2_{}'.format(mtt.DEMO_FILE)
    gh.rename_file(file_full_path_on_drive=new_file_dst, new_name=new_name)
    gh.list_all_files(starting_dir=test_f)

    # try again - will fail
    # gh.upload_content_to_new_file(dst_full_path_on_drive=new_file_dst, content='\t\thello world')
    # gh.rename_file(file_full_path_on_drive=new_file_dst, new_name=new_name)
    # gh.delete_file(full_path_on_drive=new_file_dst)

    gh.delete_file(full_path_on_drive='{}/{}'.format(test_f, new_name))
    if local_test:
        delete_test_dir(gh, test_f)
    return


def update_file_content_test(gh: got.google_handler = None, test_f: str = None):
    mt.get_function_name(ack=True, tabs=0)
    local_test = False
    if gh is None:
        gh = build_gh()
        test_f = create_test_dir(gh)
        local_test = True
    new_file_dst = '{}/{}'.format(test_f, mtt.DEMO_FILE)
    gh.upload_content_to_new_file(dst_full_path_on_drive=new_file_dst, content='\t\thello world')
    content = gh.read_file(new_file_dst)
    print(content)
    gh.update_file_content(file_full_path_on_drive=new_file_dst, new_content='\t\tgoodbye world')
    new_content = gh.read_file(new_file_dst)
    print(new_content)
    gh.delete_file(full_path_on_drive=new_file_dst)
    if local_test:
        delete_test_dir(gh, test_f)
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    mt.get_function_name(ack=True, tabs=0)
    gh = build_gh()
    new_dir_path = create_test_dir(gh)

    list_all_files_test(gh, test_f=new_dir_path)
    upload_download_delete_file_test(gh, test_f=new_dir_path)
    upload_delete_image_test(gh, test_f=new_dir_path)
    upload_read_download_delete_test(gh, test_f=new_dir_path)
    create_and_delete_empty_dir_test(gh, test_f=new_dir_path)
    create_and_delete_dir_with_files_test(gh, test_f=new_dir_path)
    download_dir_test(gh, test_f=new_dir_path)
    upload_dir_test(gh, test_f=new_dir_path)
    rename_file_test(gh, test_f=new_dir_path)
    update_file_content_test(gh, test_f=new_dir_path)

    delete_test_dir(gh, new_dir_path)
    print('{}'.format('-' * 20))
    return



import numpy as np
import os
import abc
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.socket import socket_tools as st
from wizzi_utils.open_cv import open_cv_tools as cvt
from wizzi_utils.pyplot import pyplot_tools as pyplt
from wizzi_utils.models import models_configs as cfg
# noinspection PyPackageRequirements
import cv2


class BaseModel:
    def __init__(self, save_load_dir: str, model_name: str):
        """
        :param save_load_dir: where the model is saved (or will be if not exists)
        :param model_name: valid name from models_config.py
        see:
        class_Models_test()
        """
        self.model_name = model_name
        self.local_path = save_load_dir
        if not os.path.exists(save_load_dir):
            mt.create_dir(save_load_dir, ack=False)
        return

    @staticmethod
    def model_name_valid(model_name: str) -> bool:
        if model_name not in cfg.MODELS_CONFIG:
            err = 'model {} not found in MODELS_CONFIG. options: {}'
            mt.exception_error(err.format(model_name, list(cfg.MODELS_CONFIG.keys())), real_exception=False)
            valid = False
        else:
            valid = True
        return valid

    @staticmethod
    def model_type_valid(model_type: str, model_type_needed: str) -> bool:
        if model_type not in model_type_needed:
            err = 'model type is {} but {} needed'.format(model_type, model_type_needed)
            mt.exception_error(err, real_exception=False)
            valid = False
        else:
            valid = True
        return valid

    @staticmethod
    def _download_if_needed(local_path: str, url_dict: dict) -> None:
        if not os.path.exists(local_path):
            download_style = url_dict['download_style']
            url = url_dict['url']
            if download_style == cfg.DownloadStyle.Direct.value:
                st.download_file(url=url, dst_path=local_path)
            elif download_style in [cfg.DownloadStyle.Tar.value, cfg.DownloadStyle.Zip.value]:
                root_d = os.path.dirname(local_path)
                comp_path = '{}/compressed.{}'.format(root_d, download_style)
                st.download_file(url=url, dst_path=comp_path)  # download compressed
                extracted_folder = '{}/ex'.format(root_d)
                if mt.is_windows():
                    extracted_folder = mt.full_path_no_limit(extracted_folder)

                mt.extract_file(src=comp_path, dst_folder=extracted_folder,
                                file_type=download_style)  # extract
                target_file_in_tar = url_dict['file_to_look']
                # find target file
                target_files = mt.find_files_in_folder(dir_path=extracted_folder, file_suffix=target_file_in_tar)
                if len(target_files) != 1:
                    err_msg = 'not found or found more than 1 target file in downloaded folder: {}'.format(target_files)
                    mt.exception_error(err_msg, real_exception=False)
                    exit(-1)
                mt.move_file(file_src=target_files[0], file_dst=local_path)  # move file

                mt.delete_file(file=comp_path)  # clean up tar\zip
                mt.delete_dir_with_files(dir_path=extracted_folder)  # clean up extracted_folder
        return

    @staticmethod
    def get_models_from_cfg(job: str, m_type: str = None, ack: bool = False, tabs: int = 1) -> list:
        """
        :param job: from cfg.Jobs
        :param m_type: Cv2 or Tfl or None
        :param ack:
        :param tabs:
        :return:
        """
        model_names = []
        count = 0
        for i, (m_name, m_dict) in enumerate(cfg.MODELS_CONFIG.items()):
            if m_dict['job'] == job:
                if m_type is None or (m_type is not None and m_dict['model_type'].startswith(m_type)):
                    model_names.append(m_name)
                    if ack:
                        count += 1
                        mt.dict_as_table(table=m_dict, title='{}){}'.format(count, m_name), tabs=tabs)

        return model_names

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def to_string(self, tabs: int = 1) -> str:
        print('abs method - needs implementation')
        exit(-1)
        return ''

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score percentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """
        print('abs method - needs implementation')
        exit(-1)
        return []


class OdBaseModel(BaseModel):
    DEFAULT_COLOR_DICT = {
        'label_bbox': 'black',  # if draw_labels - text bg color
        'text': 'white',  # if draw_labels - text bg color
        'sub_image': 'blue',  # if draw_sub_image - sub image bbox color
        'default_bbox': 'red',  # bbox over the detection
    }

    def __init__(
            self,
            save_load_dir: str,
            model_name: str,
            allowed_class: list = None,
    ):
        super().__init__(save_load_dir=save_load_dir, model_name=model_name)
        if not self.model_name_valid(model_name):
            exit(-1)
        self.model_cfg = cfg.MODELS_CONFIG[model_name]
        self.labels = self.model_cfg['labels_dict']['labels']
        self.allowed_classes = self.labels if allowed_class is None else allowed_class
        self.device = 'CPU'
        return

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def to_string(self, tabs: int = 1) -> str:
        print('abs method - needs implementation')
        exit(-1)
        return ''

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score percentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """
        print('abs method - needs implementation')
        exit(-1)
        return []

    @staticmethod
    def get_object_detection_models(m_type: str = None, ack: bool = False, tabs: int = 1) -> list:
        """
        :param m_type: Cv2 or Tfl or None
        :param ack:
        :param tabs:
        :return:
        """
        model_names = BaseModel.get_models_from_cfg(job=cfg.Jobs.OBJECT_DETECTION.value, m_type=m_type, ack=ack,
                                                    tabs=tabs)
        return model_names

    def draw_detections(
            self,
            detections: list,
            cv_img: np.array,
            colors_d: dict = None,
            draw_labels: bool = False,
            draw_tl_image: bool = False,
            draw_sub_image: bool = False,
            header_tl: dict = None,
            header_bl: dict = None,
            header_tr: dict = None,
            header_br: dict = None,
    ) -> None:
        """
        :param detections: output of self.classify_cv_img() on od models
        :param colors_d: colors in str form:
            bbox color
            label_bbox color
            text color
            if class_id_bbox exist - color the bbox of that class in the given color
            e.g. colors_d={
                    'label_bbox': 'black',  # if draw_labels - text bg color
                    'text': 'white',  # if draw_labels - text bg color
                    'sub_image': 'blue',  # if draw_sub_image - sub image bbox color
                    'default_bbox': 'red',  # bbox over the detection
                    'person_bbox': 'black', # custom color per class person
                    'dog_bbox': 'lime',  # custom color per class dog
                    'cat_bbox': 'magenta',  # custom color per class cat
                },
        :param cv_img: the same that was given input to self.classify_cv_img()
        :param draw_labels: draw label(label_bbox and text on it) on the bbox
        :param draw_tl_image: draw traffic light if exists - has colors in it
        :param draw_sub_image: draw sub image if exists
        :param header_tl: dict with args for cvt.add_header on top_left corner
        :param header_bl: dict with args for cvt.add_header on bottom_left corner
        :param header_tr: dict with args for cvt.add_header on top_right corner
        :param header_br: dict with args for cvt.add_header on bottom_right corner
        header e.g.
        header_tl={'text': 'wizzi_utils', 'x_offset': 0, 'text_color': 'aqua', 'with_rect': True, 'bg_color': 'black',
               'bg_font_scale': 1}
        :return:
        """
        if colors_d is None:
            colors_d = self.DEFAULT_COLOR_DICT
        for detection in detections:
            label = detection['label']
            score_percentage = detection['score_percentage']
            x0 = detection['bbox']['x0']
            y0 = detection['bbox']['y0']
            x1 = detection['bbox']['x1']
            y1 = detection['bbox']['y1']
            traffic_light_d = detection['traffic_light'] if 'traffic_light' in detection else {}
            bbox_sub_image_d = detection['bbox_sub_image'] if 'bbox_sub_image' in detection else None

            # DRAW BBOX
            k = '{}_bbox'.format(label)  # check custom color for this class
            color_bgr = pyplt.get_BGR_color(colors_d[k]) if k in colors_d else pyplt.get_BGR_color(
                colors_d['default_bbox'])
            cv2.rectangle(img=cv_img, pt1=(x0, y0), pt2=(x1, y1), color=color_bgr, thickness=2)

            if draw_tl_image:  # DRAW TL
                for loc, point_and_color in traffic_light_d.items():
                    cv2.circle(cv_img, center=tuple(point_and_color['point']), radius=2,
                               color=pyplt.get_BGR_color(point_and_color['color']), thickness=-1)

            if draw_sub_image and bbox_sub_image_d is not None:  # DRAW sub image
                pt1 = (bbox_sub_image_d['x0'], bbox_sub_image_d['y0'])
                pt2 = (bbox_sub_image_d['x1'], bbox_sub_image_d['y1'])
                cv2.rectangle(img=cv_img, pt1=pt1, pt2=pt2, color=pyplt.get_BGR_color(colors_d['sub_image']),
                              thickness=1)  # sub image

            if draw_labels:
                label_conf = '{}({}%)'.format(label, score_percentage)  # DRAW labels
                cvt.add_text(cv_img, header=label_conf, pos=(x0, y1), text_color=colors_d['text'], with_rect=True,
                             bg_color=colors_d['label_bbox'], bg_font_scale=2)

            if header_tl is not None:
                h = header_tl
                cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.TOP_LEFT.value, x_offset=h['x_offset'],
                               text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                               bg_font_scale=h['bg_font_scale'])
            if header_bl is not None:
                h = header_bl
                cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.BOTTOM_LEFT.value, x_offset=h['x_offset'],
                               text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                               bg_font_scale=h['bg_font_scale'])
            if header_tr is not None:
                h = header_tr
                cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.TOP_RIGHT.value, x_offset=h['x_offset'],
                               text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                               bg_font_scale=h['bg_font_scale'])
            if header_br is not None:
                h = header_br
                cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.BOTTOM_RIGHT.value, x_offset=h['x_offset'],
                               text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                               bg_font_scale=h['bg_font_scale'])
        return

    @staticmethod
    def add_traffic_light_to_detections(
            detections: list,
            traffic_light_p: dict,
            ack: bool = False,
            tabs: int = 1
    ) -> list:
        """
        :param detections: detections from classify_cv_img()
        :param traffic_light_p: dict of loc to float percentage

            if not none get 3 2d points in a traffic_light form
            e.g. traffic_light={
                    # x of all traffic_light points is frame_width / 2
                    'up': 0.2,  # red light will be take from y = frame_height * 0.2
                    'mid': 0.3,  # yellow light will be take from y = frame_height * 0.3
                    'down': 0.4  # green light will be take from y = frame_height * 0.3
                }
        :param ack:
        :param tabs:
        :return: for each detection in detections: add entry 'traffic_light' with a dict with keys up, mid, down
                    each has location to dict of point and color:
            e.g.
                'traffic_light'= { # x_mid = frame_width / 2
                'up': {'point': [x_mid, y_up], 'color': 'red'},  # y_up = frame_height * traffic_light['up']
                'mid': {'point': [x_mid, y_mid], 'color': 'yellow'},
                'down': {'point': [x_mid, y_down], 'color': 'green'}
                }
        """
        if ack:
            print('{}traffic light of {} detections:'.format(tabs * '\t', len(detections)))
        for detection_d in detections:
            label = detection_d['label']
            # score = detection_d['score_percentage']
            x0 = detection_d['bbox']['x0']
            y0 = detection_d['bbox']['y0']
            x1 = detection_d['bbox']['x1']
            y1 = detection_d['bbox']['y1']
            y_dist = y0 - y1  # image is flipped on y axis
            x_dist = x1 - x0
            # prepare 'traffic_light' points
            # on x middle
            # on y 20% 30% 40% from the top
            x_mid = int(x0 + 0.5 * x_dist)
            y_up = int(y0 - traffic_light_p['up'] * y_dist)
            y_mid = int(y0 - traffic_light_p['mid'] * y_dist)
            y_down = int(y0 - traffic_light_p['down'] * y_dist)
            traffic_light_out = {
                'up': {'point': [x_mid, y_up], 'color': 'red'},
                'mid': {'point': [x_mid, y_mid], 'color': 'yellow'},
                'down': {'point': [x_mid, y_down], 'color': 'green'}
            }
            detection_d['traffic_light'] = traffic_light_out
            if ack:
                print('{}\ttraffic light of detection {}: {}'.format(tabs * '\t', label, traffic_light_out))

        return detections

    @staticmethod
    def add_sub_sub_image_to_detection(
            detections: list,
            cv_img: np.array,
            bbox_image_p: dict,
            ack: bool = False,
            tabs: int = 1
    ) -> list:
        """
        :param detections:
        :param cv_img:
        :param bbox_image_p: dict that specify how much from the bbox to save
                bbox_image_p = {  # all bbox
                    'x_left_start': 0,
                    'x_left_end': 1,
                    'y_top_start': 0,
                    'y_top_end': 1,
                }
                bbox_image_p = {  # sub bbox
                    'x_left_start': 0.2,  x left: start from 20% of the image
                    'x_left_end': 0.9, x right: end at 90%
                    'y_top_start': 0.2, y top: start at 20%
                    'y_top_end': 0.6, y bottom: end at 60%
                }
        :param ack: print sub image dict
        :param tabs:
        :return: for each detection in detections: add entry 'bbox_sub_image' with a dict with keys image,x0,x1,y0,y1
                    if you imshow - you will get the bbox of the detection (according to bbox_image_p)
        """
        if ack:
            print('{}bbox_sub_image of {} detections:'.format(tabs * '\t', len(detections)))
        for detection_d in detections:
            label = detection_d['label']
            # confidence = detection_d['score_percentage']
            bbox = detection_d['bbox']
            x0, x1 = bbox['x0'], bbox['x1']
            y0, y1 = bbox['y0'], bbox['y1']

            y_dist = y1 - y0
            x_dist = x1 - x0
            # prepare sub sub image (a part of the bbox)
            # the origin is left - top. (x0,y0) is top left. (x1,y1) is bottom right
            sub_y0 = int(y0 + bbox_image_p['y_top_start'] * y_dist)
            sub_y1 = int(y0 + bbox_image_p['y_top_end'] * y_dist)
            sub_x0 = int(x0 + bbox_image_p['x_left_start'] * x_dist)
            sub_x1 = int(x0 + bbox_image_p['x_left_end'] * x_dist)

            if 0 <= sub_x0 < sub_x1 and 0 <= sub_y0 < sub_y1:
                detection_d['bbox_sub_image'] = {
                    'image': cv_img[sub_y0:sub_y1, sub_x0:sub_x1].tolist(),
                    'x0': sub_x0,
                    'y0': sub_y0,
                    'x1': sub_x1,
                    'y1': sub_y1,
                }
                # cvt.display_open_cv_image(detection_d['bbox_sub_image']['image'])
            else:
                detection_d['bbox_sub_image'] = None

            if ack:
                string = '{}\tbbox_sub_image of detection {}:'.format(tabs * '\t', label)
                bbox_sub = detection_d['bbox_sub_image']
                if bbox_sub is not None:
                    string += ' x0={} y0={} x1={} y1={} '.format(bbox_sub['x0'], bbox_sub['y0'], bbox_sub['x1'],
                                                                 bbox_sub['y1'])
                    string += '{}'.format(mt.to_str(np.array(bbox_sub['image']), title='sub_img'))
                    print(string)
                else:
                    print('{} is None'.format(string))
        return detections


class PdBaseModel(BaseModel):
    DEFAULT_COLOR_DICT = {
        'bbox_c': 'blue',
        'sub_image_c': 'darkorange',
        'text_c': 'white',
    }
    NOT_FOUND_VALUE = -1
    NOT_FOUND_PAIR = [NOT_FOUND_VALUE, NOT_FOUND_VALUE]

    def __init__(
            self,
            save_load_dir: str,
            model_name: str,
            allowed_joint_names: list = None,
            min_p_valid_joints: float = 0.3,
    ):
        super().__init__(save_load_dir=save_load_dir, model_name=model_name)
        if not self.model_name_valid(model_name):
            exit(-1)
        self.model_cfg = cfg.MODELS_CONFIG[model_name]
        self.joint_names = self.model_cfg['joint_names']
        self.pairs_indices = self.model_cfg['pairs_indices']
        # grab default colors and convert to bgr
        self.pairs_indices_colors = [pyplt.get_BGR_color(c) for c in self.model_cfg['pairs_indices_colors']]
        self.joint_colors = [pyplt.get_BGR_color(c) for c in self.model_cfg['joint_colors']]
        self.allowed_joint_names = list(
            self.joint_names.values()) if allowed_joint_names is None else allowed_joint_names
        self.min_p_valid_joints = min_p_valid_joints
        self.device = 'CPU'
        return

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def to_string(self, tabs: int = 1) -> str:
        print('abs method - needs implementation')
        exit(-1)
        return ''

    @staticmethod
    def get_pose_detection_models(m_type: str = None, ack: bool = False, tabs: int = 1) -> list:
        """
        :param m_type: Cv2 or Tfl or None
        :param ack:
        :param tabs:
        :return:
        """
        model_names = BaseModel.get_models_from_cfg(job=cfg.Jobs.POSE_DETECTION.value, m_type=m_type, ack=ack,
                                                    tabs=tabs)
        return model_names

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score percentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """
        print('abs method - needs implementation')
        exit(-1)
        return []

    def draw_detections(
            self,
            detections: list,
            cv_img: np.array,
            colors_d: dict,
            draw_joints: bool = True,
            draw_labels: bool = False,
            draw_edges: bool = False,
            draw_bbox: bool = False,
            draw_sub_image: bool = False,
            header_tl: dict = None,
            header_bl: dict = None,
            header_tr: dict = None,
            header_br: dict = None,
    ) -> None:
        """
        :param detections: output of self.classify_cv_img()
        :param colors_d: colors in str form:
            bbox_c: bbox color
            sub_image_c: sub image color
            text_c: text color
            joints_c: str or list of size self.joint_names. if None - has default in cfg
            edge_c: str or list of size self.pairs_indices. if None - has default in cfg
            e.g. colors_d={
                    'bbox_c': 'blue'
                    'sub_image_c': 'black'
                    'text_c': 'black',
                    'joints_c': str or list of size self.joint_names. if None - has default in cfg
                    'edge_c': str or list of size self.pairs_indices. if None - has default in cfg
                },
        :param cv_img: the same that was given input to self.classify_cv_img()
        :param draw_joints: draw joints
        :param draw_labels: draw joint ids (valid if draw_joints True)
        :param draw_edges: draw edges between joints
        :param draw_bbox: draw bbox around poses
        :param draw_sub_image: draw sub image from the bbox around the poses
        :param header_tl: dict with args for cvt.add_header on top_left corner
        :param header_bl: dict with args for cvt.add_header on bottom_left corner
        :param header_tr: dict with args for cvt.add_header on top_right corner
        :param header_br: dict with args for cvt.add_header on bottom_right corner
        header e.g.
        header_tl={'text': 'wizzi_utils', 'x_offset': 0, 'text_color': 'aqua', 'with_rect': True, 'bg_color': 'black',
               'bg_font_scale': 1}
        :return:
        """
        if colors_d is None:
            colors_d = self.DEFAULT_COLOR_DICT

        for detection in detections:  # each detection is a pose
            names = detection['joint_names_list']
            ids = detection['joint_ids_list']
            xys = detection['joint_x_y_list']
            # zs = detection['joint_z_list'] if 'joint_z_list' in detection else [self.NOT_FOUND_VALUE] * len(names)
            scores = detection['score_percentages_list']
            pose_bbox = detection['bbox']
            bbox_sub_image_d = detection['bbox_sub_image'] if 'bbox_sub_image' in detection else None

            if draw_bbox:  # DRAW BBOX
                color_bgr = pyplt.get_BGR_color(colors_d['bbox_c'])
                pt1 = (pose_bbox['x0'], pose_bbox['y0'])
                pt2 = (pose_bbox['x1'], pose_bbox['y1'])
                cv2.rectangle(img=cv_img, pt1=pt1, pt2=pt2, color=color_bgr, thickness=2)

            if draw_sub_image and bbox_sub_image_d is not None:  # DRAW sub image
                color_bgr = pyplt.get_BGR_color(colors_d['sub_image_c'])
                pt1 = (bbox_sub_image_d['x0'], bbox_sub_image_d['y0'])
                pt2 = (bbox_sub_image_d['x1'], bbox_sub_image_d['y1'])
                cv2.rectangle(img=cv_img, pt1=pt1, pt2=pt2, color=color_bgr, thickness=1)  # sub image

            # DRAW EDGES: lines connecting the edges
            if draw_edges:
                if 'edge_c' in colors_d:
                    edges_colors = colors_d['edge_c']
                    if mt.is_str(edges_colors):  # 1 color for all
                        edges_colors = [pyplt.get_BGR_color(edges_colors)] * len(self.pairs_indices)
                    else:  # list of colors - must be in len(self.joint_names)
                        edges_colors = [pyplt.get_BGR_color(c) for c in edges_colors]
                else:  # take default colors
                    edges_colors = self.pairs_indices_colors
                for pair, bgr_c in zip(self.pairs_indices, edges_colors):
                    partA, partB = pair
                    if xys[partA] != self.NOT_FOUND_PAIR and xys[partB] != self.NOT_FOUND_PAIR:
                        cv2.line(cv_img, pt1=tuple(xys[partA]), pt2=tuple(xys[partB]), color=bgr_c, thickness=2)

            # DRAW JOINTS: circle and joint id if asked
            if draw_joints:
                text_color_str = colors_d['text_c']
                if 'joint_colors' in colors_d:
                    joints_colors = colors_d['joint_colors']
                    if mt.is_str(joints_colors):  # 1 color for all
                        joints_colors = [pyplt.get_BGR_color(joints_colors)] * len(self.joint_names)
                    else:  # list of colors - must be in len(self.joint_names)
                        joints_colors = [pyplt.get_BGR_color(c) for c in joints_colors]
                else:  # take default colors
                    joints_colors = self.joint_colors

                for name, jid, xy, score, bgr_c in zip(names, ids, xys, scores, joints_colors):
                    if xy != self.NOT_FOUND_PAIR:
                        cv2.circle(cv_img, center=tuple(xy), radius=5, color=bgr_c, thickness=-1)
                        if draw_labels:
                            label_xy = (xy[0] - 5, xy[1])
                            label = '{}'.format(jid)
                            cvt.add_text(cv_img, header=label, pos=label_xy, text_color=text_color_str, with_rect=False,
                                         bg_font_scale=2)
            if header_tl is not None:
                h = header_tl
                cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.TOP_LEFT.value, x_offset=h['x_offset'],
                               text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                               bg_font_scale=h['bg_font_scale'])
            if header_bl is not None:
                h = header_bl
                cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.BOTTOM_LEFT.value, x_offset=h['x_offset'],
                               text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                               bg_font_scale=h['bg_font_scale'])
            if header_tr is not None:
                h = header_tr
                cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.TOP_RIGHT.value, x_offset=h['x_offset'],
                               text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                               bg_font_scale=h['bg_font_scale'])
            if header_br is not None:
                h = header_br
                cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.BOTTOM_RIGHT.value, x_offset=h['x_offset'],
                               text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                               bg_font_scale=h['bg_font_scale'])
        return

    @staticmethod
    def get_bbox_dict(joint_x_y_list: list) -> dict:
        x0, y0, x1, y1 = np.inf, np.inf, -1, -1
        for x_y in joint_x_y_list:
            if x_y != PdBaseModel.NOT_FOUND_PAIR:
                x, y = x_y
                if x0 > x:
                    x0 = x
                elif x1 < x:
                    x1 = x
                if y0 > y:
                    y0 = y
                elif y1 < y:
                    y1 = y

        # check bbox valid:
        if 0 <= x0 < x1 and 0 <= y0 < y1:
            bbox = {
                #  pt1 = (x0, y0)  # obj frame top left corner
                #  pt2 = (x1, y1)  # obj frame bottom right corner
                'x0': x0,
                'y0': y0,
                'x1': x1,
                'y1': y1,
            }
        else:
            bbox = None
        return bbox

    @staticmethod
    def add_sub_sub_image_to_detection(
            detections: list,
            cv_img: np.array,
            bbox_image_p: dict,
            ack: bool = False,
            tabs: int = 1
    ) -> list:
        """
        :param detections:
        :param cv_img:
        :param bbox_image_p: dict that specify how much from the bbox to save
                bbox_image_p = {  # all bbox
                    'x_left_start': 0,
                    'x_left_end': 1,
                    'y_top_start': 0,
                    'y_top_end': 1,
                }
                bbox_image_p = {  # sub bbox
                    'x_left_start': 0.2,  x left: start from 20% of the image
                    'x_left_end': 0.9, x right: end at 90%
                    'y_top_start': 0.2, y top: start at 20%
                    'y_top_end': 0.6, y bottom: end at 60%
                }
        :param ack: print sub image dict
        :param tabs:
        :return: for each detection in detections: add entry 'bbox_sub_image' with a dict with keys image,x0,x1,y0,y1
                    if you imshow - you will get the bbox of the detection (according to bbox_image_p)
        """
        if ack:
            print('{}bbox_sub_image of {} detections:'.format(tabs * '\t', len(detections)))
        for pose_i, detection_d in enumerate(detections):
            # label = detection_d['label']
            # confidence = detection_d['score_percentage']
            bbox = detection_d['bbox']
            x0, x1 = bbox['x0'], bbox['x1']
            y0, y1 = bbox['y0'], bbox['y1']

            y_dist = y1 - y0
            x_dist = x1 - x0
            # prepare sub sub image (a part of the bbox)
            # the origin is left - top. (x0,y0) is top left. (x1,y1) is bottom right
            sub_y0 = int(y0 + bbox_image_p['y_top_start'] * y_dist)
            sub_y1 = int(y0 + bbox_image_p['y_top_end'] * y_dist)
            sub_x0 = int(x0 + bbox_image_p['x_left_start'] * x_dist)
            sub_x1 = int(x0 + bbox_image_p['x_left_end'] * x_dist)

            if 0 <= sub_x0 < sub_x1 and 0 <= sub_y0 < sub_y1:
                detection_d['bbox_sub_image'] = {
                    'image': cv_img[sub_y0:sub_y1, sub_x0:sub_x1].tolist(),
                    'x0': sub_x0,
                    'y0': sub_y0,
                    'x1': sub_x1,
                    'y1': sub_y1,
                }
                # cvt.display_open_cv_image(detection_d['bbox_sub_image']['image'])
            else:
                detection_d['bbox_sub_image'] = None

            if ack:
                string = '{}\tbbox_sub_image of pose {}:'.format(tabs * '\t', pose_i)
                bbox_sub = detection_d['bbox_sub_image']
                if bbox_sub is not None:
                    string += ' x0={} y0={} x1={} y1={} '.format(bbox_sub['x0'], bbox_sub['y0'], bbox_sub['x1'],
                                                                 bbox_sub['y1'])
                    string += '{}'.format(mt.to_str(np.array(bbox_sub['image']), title='sub_img'))
                    print(string)
                else:
                    print('{} is None'.format(string))
        return detections



from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.misc.test import test_misc_tools as mtt
from wizzi_utils.open_cv.test import test_open_cv_tools as cvtt
from wizzi_utils.models.cv2_models import object_detection as od
from wizzi_utils.models.test.shared_code_for_tests import od_or_pd_Model_image_test
from wizzi_utils.models.test.shared_code_for_tests import od_or_pd_Model_video_test
from wizzi_utils.models.test.shared_code_for_tests import od_or_pd_Model_cam_test
from wizzi_utils.models.test.shared_code_for_tests import CAM_FRAMES, IMAGE_MS, VID_X_FRAMES_CV
from wizzi_utils.models.test.shared_code_for_tests import OD_OP, OD_IM_TEST, OD_CAM_TEST, OD_VID_TEST


def __get_od_dict(op: int = 1):
    """
    cv2 od models:
    [
        'yolov4_tiny',
        'yolov4',
        'yolov3',
        'yolov3_ssp',
        'yolo_voc',
        'MobileNetSSD_deploy',
        'yolov3_tiny',
        'yolov2_tiny_voc',
        'VGG_ILSVRC2016_SSD_300x300_deploy',
        'VGG_ILSVRC_16_layers',
        'rfcn_pascal_voc_resnet50',
        'faster_rcnn_zf',
        'ssd_inception_v2_coco_2017_11_17',
        'ssd_mobilenet_v1_coco',
        'faster_rcnn_inception_v2_coco_2018_01_28',
        'faster_rcnn_resnet50_coco_2018_01_28',
        'ssd_mobilenet_v1_coco_2017_11_17',
        'ssd_mobilenet_v1_ppn_coco',
        'ssd_mobilenet_v2_coco_2018_03_29',
        'opencv_face_detector'
    ]
    :param op:
    :return:
    """
    if op == 1:
        od_solo = {
            'model_names': ['yolov4'],
            'dis_size': (640, 480),
            'grid': (1, 1),
        }
        od_meta_dict = od_solo
    elif op == 2:
        od_dual = {
            'model_names': [
                'yolov3',
                'yolov4'
            ],
            'dis_size': (640 * 2, 480),
            'grid': (1, 2),
        }
        od_meta_dict = od_dual
    elif op == 3:
        od_all_models = {
            'model_names': od.OdBaseModel.get_object_detection_models(m_type='Cv2', ack=False),
            'dis_size': 'fs',
            'grid': (4, 5),
        }
        od_meta_dict = od_all_models
    else:
        od_meta_dict = None
    return od_meta_dict


def _get_models(model_names: list) -> list:
    models = []
    for model_name in model_names:
        m_save_dir = '{}/object_detection/{}'.format(mtt.MODELS, model_name)
        model = od.Cv2OdModel(
            save_load_dir=m_save_dir,
            model_name=model_name,
            device='gpu',
        )
        print(model.to_string(tabs=1))
        models.append(model)
    return models


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    od_meta_dict = __get_od_dict(op=OD_OP)
    models = _get_models(od_meta_dict['model_names'])
    if OD_IM_TEST:
        cv_img = cvtt.load_img_from_web(mtt.DOG, ack=False)
        od_or_pd_Model_image_test(
            cv_img=cv_img,
            models=models,
            ms=IMAGE_MS,
            dis_size=od_meta_dict['dis_size'],
            grid=od_meta_dict['grid'],
        )
    if OD_CAM_TEST:
        od_or_pd_Model_cam_test(
            models=models,
            max_frames=CAM_FRAMES,
            dis_size=od_meta_dict['dis_size'],
            grid=od_meta_dict['grid'],
        )
    if OD_VID_TEST:
        # vid_path = cvtt.get_vid_from_web(name=mtt.DOG1)
        vid_path = cvtt.get_vid_from_web(name=mtt.WOMAN_YOGA)
        od_or_pd_Model_video_test(
            models=models,
            vid_path=vid_path,
            work_every_x_frames=VID_X_FRAMES_CV,
            dis_size=od_meta_dict['dis_size'],
            grid=od_meta_dict['grid'],
        )
    print('{}'.format('-' * 20))
    return



"""
:requires: tensorflow
    gpu:
        cuda>=10: pip install tensorflow-gpu
        cuda==9.1: pip install tensorflow-gpu==1.12
        for more versions:
            https://stackoverflow.com/questions/50622525/which-tensorflow-and-cuda-version-combinations-are-compatible
    cpu: pip install tensorflow
"""
try:
    from wizzi_utils.tensorflow.tensorflow_tools import *
except ModuleNotFoundError as e:
    pass

from wizzi_utils.tensorflow import test



import numpy as np
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.open_cv import open_cv_tools as cvt
from wizzi_utils.models import models_configs as cfg
from wizzi_utils.models.base_models import OdBaseModel
# noinspection PyPackageRequirements
import cv2
# import cv3


class Cv2OdModel(OdBaseModel):
    def __init__(self,
                 save_load_dir: str,
                 model_name: str,
                 allowed_class: list = None,
                 threshold: float = None,
                 nms_threshold: float = None,
                 in_dims: tuple = None,
                 scalefactor: float = None,
                 mean: tuple = None,
                 swapRB: bool = None,
                 crop: bool = None,
                 check_type: bool = True,
                 device: str = 'cpu'
                 ):
        """
        :param save_load_dir: where the model is saved (or will be if not exists)
        :param model_name: valid name in MODEL_CONF.keys()
        :param allowed_class: ignore rest of class. list of strings
            if None - all classes allowed
        for all the following - if is None: take default value from MODELS_DNN_OBJECT_DETECTION_META_DATA['model_name']
        :param threshold: only detection above this threshold will be returned
        :param nms_threshold: non maximum suppression threshold
        :param in_dims:
        :param scalefactor:
        :param mean:
        :param swapRB:
        :param crop:
        :param check_type:
        :param device:
        example:
        model = od.Cv2OdModel(
            save_load_dir=m_save_dir,
            model_name=model_name,
            allowed_class=['dog', 'cat'],
            threshold=0.1,
            nms_threshold=0.3,
            in_dims=(416, 416),
            scalefactor=1 / 127.5,
            mean=(0, 0, 0),
            swapRB=True,
            crop=False,
        )
        see:
        best_model_images_test()
        best_model_video_test()
        best_model_cam_test()
        models_compare_images_test()
        models_compare_video_test()
        models_compare_cam_test()
        """
        super().__init__(save_load_dir=save_load_dir, model_name=model_name, allowed_class=allowed_class)
        if check_type and not self.model_type_valid(self.model_cfg['model_type'], cfg.ModelType.OdCvNormal.value):
            exit(-1)
        net = None
        if self.model_cfg['family'] == cfg.DnnFamily.Caffe.value:
            model_prototxt = "{}/{}.prototxt".format(self.local_path, self.model_name)
            model_caffe = "{}/{}.caffemodel".format(self.local_path, self.model_name)
            self._download_if_needed(local_path=model_prototxt, url_dict=self.model_cfg['prototxt'])
            self._download_if_needed(local_path=model_caffe, url_dict=self.model_cfg['caffemodel'])
            self.model_resources_sizes = [mt.file_or_folder_size(model_prototxt), mt.file_or_folder_size(model_caffe)]
            net = cv2.dnn.readNetFromCaffe(prototxt=model_prototxt, caffeModel=model_caffe)

        elif self.model_cfg['family'] == cfg.DnnFamily.Darknet.value:
            model_cfg = "{}/{}.cfg".format(self.local_path, self.model_name)
            model_weights = "{}/{}.weights".format(self.local_path, self.model_name)
            self._download_if_needed(local_path=model_cfg, url_dict=self.model_cfg['cfg'])
            self._download_if_needed(local_path=model_weights, url_dict=self.model_cfg['weights'])
            self.model_resources_sizes = [mt.file_or_folder_size(model_cfg), mt.file_or_folder_size(model_weights)]
            net = cv2.dnn.readNetFromDarknet(cfgFile=model_cfg, darknetModel=model_weights)

        elif self.model_cfg['family'] == cfg.DnnFamily.TF.value:
            model_pbtxt = "{}/{}.pbtxt".format(self.local_path, self.model_name)
            model_pb = "{}/{}.pb".format(self.local_path, self.model_name)
            self._download_if_needed(local_path=model_pbtxt, url_dict=self.model_cfg['pbtxt'])
            self._download_if_needed(local_path=model_pb, url_dict=self.model_cfg['pb'])
            self.model_resources_sizes = [mt.file_or_folder_size(model_pbtxt), mt.file_or_folder_size(model_pb)]
            net = cv2.dnn.readNetFromTensorflow(model=model_pb, config=model_pbtxt)

        if net is None:
            mt.exception_error('Failed to create network', real_exception=False)
            exit(-1)

        self.device = self.set_device(net, device)

        self.model = cv2.dnn_DetectionModel(net)

        if threshold is not None:
            self.model_cfg['threshold'] = threshold
        if nms_threshold is not None:
            self.model_cfg['nms_threshold'] = nms_threshold
        if in_dims is not None:
            self.model_cfg['in_dims'] = in_dims
        if scalefactor is not None:
            self.model_cfg['scalefactor'] = scalefactor
        if mean is not None:
            self.model_cfg['mean'] = mean
        if swapRB is not None:
            self.model_cfg['swapRB'] = swapRB
        if crop is not None:
            self.model_cfg['crop'] = crop

        # self.model.setInputParams(size=(416, 416), scale=1 / 255, swapRB=True)
        self.model.setInputParams(
            scale=self.model_cfg['scalefactor'],
            size=self.model_cfg['in_dims'],
            mean=self.model_cfg['mean'],
            swapRB=self.model_cfg['swapRB'],
            crop=self.model_cfg['crop']
        )

        return

    def to_string(self, tabs: int = 1) -> str:
        tabs_s = tabs * '\t'
        string = '{}{}'.format(tabs_s, mt.add_color(string='Cv2OdModel:', ops='underlined'))
        string += '\n\t{}name= {} (resources size: {})'.format(tabs_s, self.model_name, self.model_resources_sizes)
        string += '\n\t{}Running on {}'.format(tabs_s, self.device)
        string += '\n\t{}local_path={}'.format(tabs_s, self.local_path)
        string += '\n\t{}{}'.format(tabs_s, mt.to_str(self.allowed_classes, 'allowed_class'))
        string += '\n{}'.format(mt.dict_as_table(self.model_cfg, title='conf', fp=6, ack=False, tabs=tabs + 1))
        return string

    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score precentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """
        classes, scores, boxes = self.model.detect(cv_img, self.model_cfg['threshold'], self.model_cfg['nms_threshold'])
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(classes, 'classes')))
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(scores, 'scores')))
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(boxes, 'boxes')))
        if len(classes) > 0 and len(scores) > 0:
            classes = classes.flatten()
            scores = scores.flatten()

        detections = self.extract_results(
            classes=classes,
            scores=scores,
            boxes=boxes,
            cv_img=cv_img,
            fp=fp,
            ack=ack,
            tabs=tabs,
            img_title=img_title
        )

        return detections

    def extract_results(
            self,
            classes: np.array,
            scores: np.array,
            boxes: np.array,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param classes: result of classes, scores, boxes = self.model.detect(cv_img, self.threshold, self.nms_threshold)
        :param scores: result of classes, scores, boxes = self.model.detect(cv_img, self.threshold, self.nms_threshold)
        :param boxes: result of classes, scores, boxes = self.model.detect(cv_img, self.threshold, self.nms_threshold)
        :param cv_img: cv image
        :param fp: float precision on the score precentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts:
        dict is a detection of an object above threshold.
            has items:
                label:str e.g. 'person'
                score_percentage: float e.g. 12.31
                bbox: dict with keys x0,y0,x1,y1
                #  pt1 = (x0, y0)  # obj frame top left corner
                #  pt2 = (x1, y1)  # obj frame bottom right corner
        """

        if ack:
            title_suffix = '' if img_title is None else '{} '.format(img_title)
            title = '{} detections({}) on image {}{}:'.format(self.model_name, len(classes), title_suffix, cv_img.shape)

            print('{}{}'.format(tabs * '\t', title))
            print('{}Meta_data:'.format(tabs * '\t'))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(classes, 'classes')))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(scores, 'scores')))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(boxes, 'boxes')))
            print('{}Detections:'.format(tabs * '\t'))

        img_h, img_w = cv_img.shape[0], cv_img.shape[1]
        detections = []
        for i, (class_id, score_frac, bbox) in enumerate(zip(classes, scores, boxes)):
            label = self.labels[class_id]
            if label in self.allowed_classes:
                score_percentage = round(score_frac * 100, fp)

                (x0, y0) = (bbox[0], bbox[1])
                (w, h) = (bbox[2], bbox[3])

                x1 = min(x0 + w, img_w)  # dont exceed frame width
                y1 = min(y0 + h, img_h)  # dont exceed frame height

                if 'need_normalize' in self.model_cfg:
                    in_dims_w, in_dims_h = self.model_cfg['in_dims']
                    x0 = int(x0 * (img_w / in_dims_w))
                    x1 = int(x1 * (img_w / in_dims_w))
                    y0 = int(y0 * (img_h / in_dims_h))
                    y1 = int(y1 * (img_h / in_dims_h))

                detection_d = {
                    'label': label,
                    'score_percentage': score_percentage,
                    'bbox': {
                        #  pt1 = (x0, y0)  # obj frame top left corner
                        #  pt2 = (x1, y1)  # obj frame bottom right corner
                        'x0': int(x0),
                        'y0': int(y0),
                        'x1': int(x1),
                        'y1': int(y1),
                    },
                }

                detections.append(detection_d)
                if ack:
                    d_msg = '{}\t{})Detected {}({}%) in top left=({}), bottom right=({})'
                    print(d_msg.format(tabs * '\t', i, label, score_percentage, (x0, y0), (x1, y1)))

        return detections

    @staticmethod
    def set_device(net: cv2.dnn, device: str = 'cpu') -> str:
        string = 'CPU'
        if device == "gpu":
            if cvt.cuda_on():
                net.setPreferableBackend(cv2.dnn.DNN_BACKEND_CUDA)
                net.setPreferableTarget(cv2.dnn.DNN_TARGET_CUDA)
                string = 'GPU'
            else:
                mt.exception_error('GPU requested but open_cv can\'t find cuda device')
                net.setPreferableBackend(cv2.dnn.DNN_TARGET_CPU)
        else:  # device == "cpu":
            net.setPreferableBackend(cv2.dnn.DNN_TARGET_CPU)
        return string



import numpy as np
import os
import copy
from wizzi_utils import misc_tools as mt  # misc tools

# noinspection PyPackageRequirements
import torch
# noinspection PyPackageRequirements
import torch.nn as nn
# noinspection PyPackageRequirements
import torchvision

# found that these are the most used transforms for mnist and cifar10
DEFAULT_MNIST_TR = torchvision.transforms.Compose([
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize((0.1307,), (0.3081,))])

DEFAULT_CIFAR10_TR_TRAIN = torchvision.transforms.Compose([
    torchvision.transforms.Pad(4),
    torchvision.transforms.RandomCrop(32),
    torchvision.transforms.RandomHorizontalFlip(),
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
])

DEFAULT_CIFAR10_TR_TEST = torchvision.transforms.Compose([
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)), ])


def cuda_on() -> bool:
    """
     check if cuda available
     see cuda_on_test()
     """
    return torch.cuda.is_available()


def set_cuda_scope_and_seed(seed: int, dtype='FloatTensor', tabs: int = 1) -> None:
    """
    :param seed: setting torch seed and default torch if cuda on
    :param dtype:
        https://pytorch.org/docs/stable/tensors.html
        32-bit floating point: torch.cuda.FloatTensor
        64-bit floating point: torch.cuda.DoubleTensor
    :param tabs:
    see set_cuda_scope_and_seed_test()
    """
    mt.set_seed(seed)
    if cuda_on():
        def_dtype = 'torch.cuda.{}'.format(dtype)
        torch.set_default_tensor_type(def_dtype)
        torch.manual_seed(seed)
        # torch.cuda.manual_seed(seed)
        print('{}working on CUDA. default dtype = {} <=> {}'.format(tabs * '\t', def_dtype, torch.get_default_dtype()))
    else:
        torch.manual_seed(seed)
        print('{}working on CPU'.format(tabs * '\t'))
    return


def add_cuda(var: torch.Tensor) -> torch.Tensor:
    """
    assigns the variables to GPU if available
    see add_cuda_test()
    """
    if cuda_on() and not is_cuda(var):
        var = var.cuda()
    return var


def is_trainable(var: torch.Tensor) -> bool:
    """
    :param var:
    :return: if the variable is trainable
    see is_trainable_test()
    """
    return var.requires_grad


def set_trainable(var: torch.Tensor, trainable: bool = True) -> None:
    """
    :param var:
    :param trainable: True\False
    sets the var to 'trainable'
    """
    var.requires_grad = trainable
    return


def is_cuda(var: torch.Tensor) -> bool:
    """
    :param var:
    :return: if the variable is Cuda\CPU variable
    see is_cuda_test()
    """
    # noinspection PyTypeChecker
    return var.is_cuda


def size_s(var: torch.Tensor) -> str:
    """
    :param var:
    :return: clean str of tensor size
    e.g. torch.Size([1, 3, 29]) -> [1, 3, 29]
    see size_s_test()
    """
    size_str = str(var.size())
    size_str = size_str[size_str.find("(") + 1:size_str.find(")")]
    return size_str


def total_size(t: torch.Tensor, ignore_first=True) -> int:
    """
    :param t:
    :param ignore_first: if the first dim is the rows and you want each row size
    calculates the total size of a tensor.
    see total_size_test()
    """
    total = 1
    my_shape = t.shape[1:] if ignore_first else t.shape
    for d in my_shape:
        total *= d
    return total


def torch_to_numpy(var_torch: torch.Tensor) -> np.array:
    """
    :param var_torch:
    :return: np array of var_torch
    convert torch to numpy
    see torch_to_numpy_test()
    """
    if is_trainable(var_torch):
        var_np = var_torch.detach().cpu().numpy()
    else:
        var_np = var_torch.cpu().numpy()
    return var_np


def numpy_to_torch(var_np: np.array, to_double=True, detach: bool = False) -> torch.Tensor:
    """
    :param var_np:
    :param to_double: create torch double - else float
    :param detach: if needed to be detached - don't remember when detached needed to be true happened
    convert numpy to torch also add to cude if cuda is on
    see numpy_to_torch_test()
    """
    if detach:
        if to_double:
            var_torch = add_cuda(torch.from_numpy(var_np).double()).detach()
        else:
            var_torch = add_cuda(torch.from_numpy(var_np).float()).detach()
    else:
        if to_double:
            var_torch = add_cuda(torch.from_numpy(var_np).double())
        else:
            var_torch = add_cuda(torch.from_numpy(var_np).float())
    return var_torch


def to_str(var: any,
           title: str = 'var',
           chars: int = 100,
           fp: int = 2,
           wm: bool = True,
           rec: bool = False
           ) -> str:
    """
    :param var: the variable
    :param title: str: the title (usually variable name)
    :param chars: int, None or str:
        chars>0: maximal number of chars
        chars==None: no chars
        chars=='all': all chars
    :param fp: float_precision: round number if possible(float, list or np array of floats...)
            fp>0 round
            fp==None: no rounding
    :param wm: with_meta: with meta data such as type, len\shape, dtype...
    :param rec: recursive: to keep printing if there are more items inside e.g. np.array(shape=(2,3,4)) -> 3 prints
    :return: informative string of the variable
    see to_str_test() on test_torch_tools.py
    """

    string = title
    type_s = str(type(var)).replace('<class \'', '').replace('\'>', '')  # clean type name

    if is_torch(var):
        if wm:
            string += '({}'.format(type_s)
            string += ',s={}'.format(var.shape)
            string += ',dtype={}'.format(var.dtype)
            string += ',trainable={}'.format(is_trainable(var))
            string += ',is_cuda={})'.format(is_cuda(var))

        if chars != 0:  # -1 for all, or x>0 for x chars
            # new_v = var.tolist()
            if fp >= 0 and len(var.size()) > 0:
                if all((mt.is_int(item) or mt.is_float(item)) for item in var):  # 1d list of int and floats
                    if all(mt.is_int(item) for item in var):  # if all ints - no rounding
                        f_format = '{:,}'
                    else:
                        f_format = '{:,.%xf}' % fp
                    new_v = [f_format.format(li_item) for li_item in var]
                else:  # >1d list or 1d with not just ints and floats
                    new_v = np.around(var.tolist(), fp).tolist()
            else:
                new_v = var
            data_str_raw = str(new_v).replace('\'', '')
            string += mt.get_data_str(data_str_raw, chars)

        if rec and len(var.size()) > 0:  # recursive call
            inner_str = to_str(var=var[0], title='{}[0]'.format(title), chars=chars, fp=fp, rec=rec)
            string += '\n\t{}'.format(inner_str)
    else:  # if it's not Tensor - use the default to_str() from misc_tools
        string = mt.to_str(var, title=title, chars=chars, wm=wm, fp=fp, rec=rec)
    return string


def save_tensor(t, path: str, ack_print: bool = True, tabs: int = 1):
    """
    :param t: dict or a tensor
    :param path:
    :param ack_print:
    :param tabs:

    see save_load_tensor_test()
    """
    torch.save(t, path)
    if ack_print:
        size_str = mt.file_or_folder_size(path)
        file_msg = '{}({})'.format(path, size_str)
        print('{}{}'.format(tabs * '\t', mt.SAVED.format(file_msg)))
    return


def load_tensor(path: str, ack_print: bool = True, tabs: int = 1):
    """
    :param path:
    :param ack_print:
    :param tabs:
    :return: t: dict or a tensor
    see save_load_tensor_test()
    """
    if os.path.exists(path):
        dst_allocation = None if cuda_on() else 'cpu'
        t = torch.load(path, map_location=dst_allocation)
        if ack_print:
            size_str = mt.file_or_folder_size(path)
            file_msg = '{}({})'.format(path, size_str)
            print('{}{}'.format(tabs * '\t', mt.LOADED.format(file_msg)))
    else:
        mt.exception_error(mt.NOT_FOUND.format(path), real_exception=False, tabs=tabs)
        t = None
    return t


def torch_uniform(shape: tuple, range_low: float, range_high: float) -> torch.Tensor:
    """
    :param shape:
    :param range_low:
    :param range_high:
    :return:
    see torch_uniform_test()
    """
    ret = torch.empty(shape).uniform_(range_low, range_high)
    return ret


def torch_normal(shape: tuple, miu: float, std: float) -> torch.Tensor:
    """
    :param shape:
    :param miu:
    :param std:
    :return:
    see torch_normal_test()
    """
    ret = torch.empty(shape).normal_(miu, std)
    return ret


def opt_to_str(optimizer: torch.optim) -> str:
    """
    :param optimizer:
    :return:
    see opt_to_str_test()
    """
    opt_s = str(optimizer).replace('\n', '').replace('    ', ' ')
    return opt_s


def get_lr(optimizer: [torch.optim, int]) -> float:
    """
    :param optimizer:
    :return:
    see get_lr_test()
    """
    lr = None
    if isinstance(optimizer, OptimizerHandler):
        lr = optimizer.lr()
    else:  # isinstance(optimizer, torch.Tensor):
        # noinspection PyUnresolvedReferences
        if "lr" in optimizer.param_groups[0]:
            # noinspection PyUnresolvedReferences
            lr = optimizer.param_groups[0]['lr']
    return lr


def set_lr(optimizer: torch.optim, new_lr: float) -> None:
    """
    :param optimizer:
    :param new_lr:
    :return:
    see set_lr_test()
    """
    optimizer.param_groups[0]['lr'] = new_lr
    return


def get_opt_by_name(opt_d: dict, params: list) -> torch.optim:
    """
    :param params: trainable params
    :param opt_d: has keys: name, lr, weight_decay and optional if SGD momentum
        e.g. options: {'name'= 'ADAM', 'lr': 0.001, 'weight_decay': 0}
    :return: optimizer
    see get_opt_by_name_test()
    """
    opt = None
    if opt_d['name'] == 'ADAM':
        opt = torch.optim.Adam(params, lr=opt_d['lr'], weight_decay=opt_d['weight_decay'])
    elif opt_d['name'] == 'SGD':
        opt = torch.optim.SGD(params, lr=opt_d['lr'], momentum=opt_d['momentum'], weight_decay=opt_d['weight_decay'])
    return opt


class OptimizerHandler:
    """
    Optimizer wrapper.
    if update_lr() called 'patience' times:
        new_lr = max(factor * lr, min_lr)
        set_lr(opt, new_lr)
    see OptimizerHandler_test()
    """

    def __init__(self, optimizer: torch.optim, factor: float, patience: int, min_lr: float):
        """
        :param optimizer: class torch.optim: e.g. Adam, SGD ...
        :param factor: percent to keep from old lr. if factor 0.1 and lr was 5: new lr = 0.5
        :param patience: how long to wait before changing lr
        :param min_lr: minimal lr
        """
        self.optimizer = optimizer
        self.factor = factor
        self.patience = patience
        self.min_lr = min_lr
        self.epochs_passed = 0

    def step(self):
        self.optimizer.step()
        return

    def lr(self):
        return self.optimizer.param_groups[0]['lr']

    def set_lr(self, new_lr: float):
        self.optimizer.param_groups[0]['lr'] = new_lr
        return

    def zero_grad(self):
        self.optimizer.zero_grad()
        return

    def update_lr(self):
        self.epochs_passed += 1
        if self.epochs_passed >= self.patience:
            self.epochs_passed = 0
            old_lr = self.lr()
            new_lr = max(old_lr * self.factor, self.min_lr)
            self.set_lr(new_lr)
            # print('new lr changed to {}'.format(self.lr()))
        return


class EarlyStopping:
    """
    counts how many rounds no improvment in loss
    if that passed patience - return true in should_early_stop()
    see EarlyStopping_test()
    """

    def __init__(self, patience: int):
        """
        :param patience:
        """
        self.patience = patience
        self.counter = 0
        self.best = None
        return

    def should_early_stop(self, loss: float):
        """
        :param loss:
        :return:
        """
        should_stop = False
        if self.best is None:
            self.best = loss
        elif loss < self.best:
            self.best = loss
            self.counter = 0
        else:
            self.counter += 1
            # print('\t\tpatience {}/{}'.format(self.counter, self.patience))
            if self.counter >= self.patience:
                should_stop = True
        return should_stop


def subset_init(c_size: int, A: torch.Tensor, trainable: bool = True) -> torch.Tensor:
    """
    :param c_size:
    :param A:
    :param trainable:
    :return:
    given c_size <= |A|, initialize a tensor C with a random subset of A
    see subset_init_test()
    """
    n = A.shape[0]
    if c_size >= n:
        if A.dtype == 'torch.float64':
            C = A.clone().double()
        elif A.dtype == 'torch.float32':
            C = A.clone().float()
        else:
            C = A.clone()
    else:
        perm = np.random.permutation(n)
        idx = perm[:c_size]
        if A.dtype == 'torch.float64':
            C = A[idx].clone().double()
        elif A.dtype == 'torch.float32':
            C = A[idx].clone().float()
        else:
            C = A[idx].clone()

    set_trainable(C, trainable)
    return C


def augment_x_y_torch(X: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    :param X:
    :param y:
    :return:
    creates A=X|y
    see augment_x_y_torch_test()
    """
    assert X.shape[0] == y.shape[0], 'row count must be the same'
    if len(X.shape) == 1:  # change x size()=[n,] to size()=[n,1]
        X = X.view(X.shape[0], 1)
    if len(y.shape) == 1:  # change y size()=[n,] to size()=[n,1]
        y = y.view(y.shape[0], 1)
    A = torch.cat((X, y), 1)
    return A


def de_augment_torch(A: torch.Tensor) -> (torch.Tensor, torch.Tensor):
    """
    :param A:
    :return:
    creates X|y=A
    see de_augment_torch_test()
    """
    assert 0 < len(A.shape) <= 2, 'supports 2d only'
    if len(A.shape) == 1:  # A is 1 point. change from size (n) to size (1,n)
        A = A.view(1, A.shape[0])
    X, y = A[:, :-1], A[:, -1]
    if len(X.shape) == 1:  # change x size()=[n,] to size()=[n,1]
        X = X.view(X.shape[0], 1)
    if len(y.shape) == 1:  # change y size()=[n,] to size()=[n,1]
        y = y.view(y.shape[0], 1)
    return X, y


def split_tensor(Q: torch.Tensor, p: float = 0.9) -> (torch.Tensor, torch.Tensor):
    """
    :param Q:
    :param p:
    :return:
    see split_tensor_test()
    """
    partition = int(p * Q.shape[0])
    Q_1 = Q[:partition]
    Q_2 = Q[partition:]
    return Q_1, Q_2


def shuffle_tensor(arr: torch.Tensor) -> torch.Tensor:
    """
    :param arr:
    :return:
    shuffles an array
    see shuffle_tensor_test()
    """
    if is_torch(arr):
        arr = arr[torch.randperm(arr.shape[0])]
    return arr


def shuffle_tensors(arr_tuple: tuple) -> tuple:
    """
    :param arr_tuple: tuple of arrays (numpy or tensor)
        len(arr) is equal on all arrays
    :return: shuffled arrays
    see shuffle_tensors_test()
    """
    arrays_size = len(arr_tuple[0])
    # rand_perm = torch.randperm(arrays_size)
    rand_perm = np.random.permutation(arrays_size)

    out_tuple = ()
    for arr in arr_tuple:
        arr_shf = arr[rand_perm]
        out_tuple += (arr_shf,)
    return out_tuple


def is_torch(var: any) -> bool:
    # noinspection PyTypeChecker
    return isinstance(var, torch.Tensor)


def count_keys(y: [torch.Tensor, np.array, list], tabs: int = 1) -> None:
    """
    :param y: nx1 array (torch, list, numpy)
    :param tabs:
    see count_keys_test()
    """
    from collections import Counter
    if hasattr(y, "shape"):
        y_shape = y.shape
    else:
        y_shape = len(y)
    print('{}Count classes: (y shape {})'.format(tabs * '\t', y_shape))
    cnt = Counter()
    for value in y:
        ind = value.item() if is_torch(y) else value
        cnt[ind] += 1
    cnt = sorted(cnt.items())
    for item in cnt:
        print('{}\tClass {}: {} samples'.format(tabs * '\t', item[0], item[1]))
    return


def get_torch_version(ack: bool = False, tabs: int = 1) -> str:
    """
    :return:
    see get_torch_version_test()
    """
    # noinspection PyUnresolvedReferences
    string = mt.add_color('{}* PyTorch Version {}'.format(tabs * '\t', torch.__version__), ops=mt.SUCCESS_C)
    string += mt.add_color(' - GPU detected ? ', ops=mt.SUCCESS_C)
    if cuda_on():
        string += mt.add_color('True', ops=mt.SUCCESS_C2)
    else:
        string += mt.add_color('False', ops=mt.FAIL_C2)
    if ack:
        print(string)
    return string


def load_model(model: nn.Module, path: str, ack: bool = True, tabs: int = 1) -> None:
    """
    :param model: nn model
    :param path: checkpoint path - ends with .pt or .pth
    :param ack:
    :param tabs:
    :return: None - by reference
    see save_load_model_test()
    """
    if os.path.exists(path):
        dst_allocation = None if torch.cuda.is_available() else 'cpu'
        model.load_state_dict(torch.load(path, map_location=dst_allocation))
        model.eval()
        if ack:
            size_str = mt.file_or_folder_size(path)
            file_msg = '{}({})'.format(path, size_str)
            print('{}{}'.format(tabs * '\t', mt.LOADED.format(file_msg)))
    else:
        mt.exception_error(mt.NOT_FOUND.format(path), real_exception=False, tabs=tabs)
    return


def save_model(model: nn.Module, path: str, ack: bool = True, tabs: int = 1) -> None:
    """
    :param model: nn model to save
    :param path: checkpoint path - ends with .pt or .pth
    :param ack:
    :param tabs:
    :return:
    see save_load_model_test()
    """
    torch.save(model.state_dict(), path)
    if ack:
        size_str = mt.file_or_folder_size(path)
        file_msg = '{}({})'.format(path, size_str)
        print('{}{}'.format(tabs * '\t', mt.SAVED.format(file_msg)))
    return


def model_info(model: nn.Module, title: str = 'model', wv: bool = False, chars: int = 200, tabs: int = 1):
    """
    :param model: nn model with self.title member
    :param title:
    :param wv: with values: print variables values
    :param chars: if wv: prints first 'chars' for each variable
    :param tabs:
    :return:
    see model_info_test()
    """
    print('{}{}:'.format(tabs * '\t', title))
    var_meta = '{}\t{}: {}({:,.0f} params), dtype:{}, trainable:{}, is_cuda:{}'
    sum_params = 0
    for name, param in model.named_parameters():
        layer_params = total_size(param, ignore_first=False)
        sum_params += layer_params
        print(var_meta.format(tabs * '\t', name, size_s(param), layer_params, param.dtype, is_trainable(param),
                              is_cuda(param)))
        # print('{}\t{}'.format(tabs * '\t', to_str(var=param, title=name)))
        if wv:
            print('{}\t\t{}'.format(tabs * '\t', mt.to_str(param.tolist(), 'values', wm=False, chars=chars)))
    print('{}\tTotal {:,.0f} params'.format(tabs * '\t', sum_params))
    return


def layer_info(model: nn.Module, layer: str, title: str = 'layer', wv: bool = False, chars: int = 200, tabs: int = 1):
    """
    :param model: nn model with self.title member
    :param layer: layer partial name
        e.g. 'conv2' -> params conv2.weight conv2.bias will be printed
    :param title:
    :param wv: with values: print variables values
    :param chars: if wv: prints first 'chars' for each variable
    :param tabs:
    :return:
    see layer_info_test()
    """
    print('{}{}:'.format(tabs * '\t', title))
    var_meta = '{}\t{}: {}({:,.0f} params), dtype:{}, trainable:{}, is_cuda:{}'
    for name, param in model.named_parameters():
        if layer in name:
            layer_params = total_size(param, ignore_first=False)
            print(var_meta.format(tabs * '\t', name, size_s(param), layer_params, param.dtype, is_trainable(param),
                                  is_cuda(param)))
            if wv:
                print('{}\t\t{}'.format(tabs * '\t', mt.to_str(param.tolist(), 'values', wm=False, chars=chars)))
    return


def model_summary(model: nn.Module, input_size: tuple, batch_size: int) -> None:
    """
    prints model summary
    # if cuda_on(): model_summary need float variables
    e.g.
        m = MnistModel().float()
        tt.model_summary(m, (1, 28, 28), 64))
    see model_summary_test()
    """
    try:
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from torchsummary import summary
        summary(model, input_size, batch_size)
    except (ModuleNotFoundError, ImportError) as e:
        mt.exception_error(e, real_exception=True)
    return


def model_params_count(model: nn.Module) -> int:
    """
    :param model:
    :return: counts parameters
    see model_params_count_test()
    """
    total_parameters = 0
    for p in list(model.parameters()):
        total_parameters += total_size(p, False)
    return total_parameters


def model_status_str(model: nn.Module) -> str:
    """
    3 options: model fully trainable, fully frozen, both
    see model_status_str_test()
    """
    saw_trainable, saw_frozen = False, False
    for param in model.parameters():
        if is_trainable(param):
            saw_trainable = True
        else:
            saw_frozen = True

    if saw_frozen and saw_trainable:
        msg = 'partly trainable and partly frozen'
    elif saw_trainable:
        msg = 'fully trainable'
    else:
        msg = 'fully frozen'
    return msg


def set_model_status(model: nn.Module, status: bool) -> None:
    """
    :param model:
    :param status:
    :return:
    set model parameters trainable status to 'status'
    see set_model_status_test()
    """
    for param in model.parameters():
        set_trainable(param, status)
    return


def set_layer_status(model: nn.Module, layers: list, status: bool) -> None:
    """
    :param model:
    :param layers: part of the name of the layer in the model.
        e.g. 'conv2' -> params conv2.weight conv2.bias will change trainable status to `status`
    :param status:
    :return:

    see set_layer_status_test()
    """
    for name, param in model.named_parameters():
        for layer in layers:
            if layer in name:
                param.requires_grad_(status)
                break
    return


def model_clone(model: nn.Module) -> nn.Module:
    """
    :param model: deep copy model
    :return:
    see model_clone_test()
    """
    m_clone = copy.deepcopy(model)
    return m_clone


def model_copy_layers(model_src: nn.Module, model_dst: nn.Module, layers: list) -> None:
    """
    :param model_src: assuming same size as model_dst
    :param model_dst: assuming same size as model_src
    :param layers:
    :return:
    see model_copy_layers_test()
    """
    for name, param in model_src.named_parameters():
        current_should_be_copied = any(layer in name for layer in layers)
        if current_should_be_copied:
            model_dst.state_dict()[name].copy_(param)
    return


def model_copy_except_layers(model_src: nn.Module, model_dst: nn.Module, layers: list) -> None:
    """
    :param model_src:
    :param model_dst:
    :param layers:
    :return:
    see model_copy_layer_test()
    """
    for name, param in model_src.named_parameters():
        current_should_be_copied = all(layer not in name for layer in layers)
        if current_should_be_copied:
            model_dst.state_dict()[name].copy_(param)
    return


def size_s_ds(ds: torchvision.datasets) -> str:
    """
    :param ds:
    :return:
    see size_s_ds_test()
    """
    ds_len = len(ds)

    x = ds.data[0]  # load 1st sample as data loader will load
    X_size_post_tr = (ds_len,)
    for d in x.shape:
        X_size_post_tr += (d,)

    y_size = (len(ds.targets),)  # real data
    res = '|X|={}, |y|={}'.format(X_size_post_tr, y_size)
    return res


def get_data_set(
        ds_name: str,
        data_root: str,
        is_train: bool,
        transform=None,
        download: bool = False,
        data_limit: int = None,
        ack: bool = True,
        tabs: int = 1
) -> torchvision.datasets:
    """
    :param ds_name: 'mnist' of 'cifar10'
    :param data_root: folder containing the ds.
        if not containing and download == True: download to data_root
    :param is_train: train(True) or test
    :param transform: transform to images
    :param download: if to download if not found
    :param data_limit: good for debugging - shrinking the dataset from X to Y s.t. Y<X
    :param ack:
    :param tabs:
    :return:
    """
    if ds_name == 'cifar10':
        if transform is None:  # default cifar10 transforms
            transform = DEFAULT_CIFAR10_TR_TRAIN if is_train else DEFAULT_CIFAR10_TR_TEST
        data_set = torchvision.datasets.CIFAR10(root=data_root, train=is_train, download=download, transform=transform)
    elif ds_name == 'mnist':
        if transform is None:  # default mnist transform
            transform = DEFAULT_MNIST_TR
        data_set = torchvision.datasets.MNIST(root=data_root, train=is_train, download=download, transform=transform)
    else:
        mt.exception_error('data_set not valid!', real_exception=False)
        return
    if data_limit is not None:
        data_set.data = data_set.data[:data_limit]
        data_set.targets = data_set.targets[:data_limit]

    if ack:
        prefix = 'train' if is_train else 'test'
        print('{}{}({}) dataset: {}'.format(tabs * '\t', ds_name, prefix, size_s_ds(data_set)))
    return data_set



from wizzi_utils.algorithms import algorithms as alg
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.pyplot import pyplot_tools as pyplt
import numpy as np

PLT_MAX_TIME = 2  # close plt after x seconds - minimum 2


def find_centers_test():
    """
    if you get No module named 'sklearn.__check_build._check_build'
    pip uninstall scikit-learn
    pip uninstall sklearn
    pip install sklearn
    :return:
    """
    mt.get_function_name(ack=True, tabs=0)
    A = np.array([[-1, -1], [-1, 1], [1, 1], [1, -1]])
    centers = alg.find_centers(A, k=1)
    if centers is not None:
        print('\tVisual test: square and it\'s center')

        pyplt.plot_2d_one_figure(
            datum=[
                {
                    'data': A,
                    'c': 'g',
                    'label': 'Data',
                    'marker': 'o',
                },
                {
                    'data': centers,
                    'c': 'r',
                    'label': '1Mean',
                    'marker': 'x',
                }
            ],
            fig_title='square and it\'s center',
            win_d={
                'title': 'find_centers_test()',
                'location': (0, 0),
                'resize': 1,
                'zoomed': False
            },
            max_time=PLT_MAX_TIME
        )
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    find_centers_test()
    print('{}'.format('-' * 20))
    return



from wizzi_utils.misc import misc_tools as mt
import os
# noinspection PyPackageRequirements
import yaml
# noinspection PyPackageRequirements
from pydrive.auth import GoogleAuth
# noinspection PyPackageRequirements
from pydrive.auth import RefreshError
# noinspection PyPackageRequirements
from pydrive.drive import GoogleDrive
# noinspection PyPackageRequirements
from pydrive.files import GoogleDriveFile
# noinspection PyPackageRequirements
from pydrive.settings import InvalidConfigError


def read_yaml_to_dict(yaml_path: str) -> dict:
    """
    :param yaml_path:
    :return:
    """
    if os.path.exists(yaml_path):
        with open(yaml_path) as f:
            yaml_d = yaml.load(f, Loader=yaml.FullLoader)
    else:
        mt.exception_error(mt.NOT_FOUND.format(yaml_path), real_exception=False, tabs=1)
        yaml_d = None
    return yaml_d


class google_handler:
    """
    Before starting - know i made an assumption that a file\dir name is unique.
    From some reason, google allow same file or dir name in the same dir. e.g. 'root/a.txt' twice
    If you have duplicates - this will be notified upon creation of this class

    api: https://pythonhosted.org/PyDrive/quickstart.html
    example: https://www.dezyre.com/recipes/upload-files-to-google-drive-using-python
    instructions:
        HOST: google account that handles the project
        CLIENT: google account of any user that want to use the project above
        * HOST could be a CLIENT
     HOST:
         1 https://console.cloud.google.com/
         2 Dashboard - create project
         3 enter project - enable Api and services - type google drive and enable
         4 CREATE CREDENTIALS - client id - desktop app and download the dot json file -
            can rename to client_secret.json
     CLIENT:
         1 HOST must add CLIENT (including himself):
             in OAuth consent screen: +Add users, under test users, add user email
         2 pip install pydrive
             see more at https://pythonhosted.org/PyDrive/
         3 create dir for the client connection e.g. GDrive2
             place in it the client_secret.json file from HOST step 4
             create file settings.yaml with content:
                 client_config_backend: file
                 client_config_file: ./GDrive2/client_secret.json
                 save_credentials: True
                 save_credentials_backend: file
                 save_credentials_file: ./GDrive2/credentials.json
                 oauth_scope: ['https://www.googleapis.com/auth/drive']
         4 on first run - it will open a browser and ask to enter user and pass. user must be added on CLIENT step 1
             after first run credentials will be created in the path stated on the settings.yaml save_credentials_file
             no more login required
    """

    # noinspection PyPackageRequirements
    def __init__(self,
                 yaml_path: str,
                 tabs: int = 1,
                 dir_color: (str, list) = 'reverse',
                 file_color: (str, list) = 'blue',
                 title_color: (str, list) = 'underlined'
                 ):
        if not os.path.exists(yaml_path):
            mt.exception_error('cant find {}'.format(yaml_path), real_exception=False, tabs=0)
            print('File name settings.yaml should look like(you can copy paste):')
            print('\tclient_config_backend: file')
            print('\tclient_config_file: ./GDrive2/client_secret.json')
            print('\tsave_credentials: True')
            print('\tsave_credentials_backend: file')
            print('\tsave_credentials_file: ./GDrive2/credentials.json')
            print('\toauth_scope: [\'https://www.googleapis.com/auth/drive\']')
            print('*** assumes you created a folder named GDrive2 in current dir')
            print('*** assumes your client_secret HOST downloaded and gave you in GDrive2 named client_secret.json')
            print('*** the credentials_file will be saved after 1st run in GDrive2 under name credentials.json')
            exit(-1)
        self.yaml_path = yaml_path
        self.yaml_d = read_yaml_to_dict(self.yaml_path)
        try:
            # assuming we have valid client secret file and a settings yaml,
            # upon connection there are 3 options:
            # 1 no cred file - easy - browser opens to log in
            # 2 valid cred file - easy - connection made and no need for anything more
            # 3 expired cred file - rename the current cred file and try connecting agian to log in
            self.all_files_dict = None
            self.gauth = GoogleAuth(settings_file=yaml_path)
            self.drive = GoogleDrive(self.gauth)
            try:  # option 3
                self.__refresh_all_files_list()  # updates self.all_files_list
                # if expired cred file - this will raise refresh error
            except RefreshError as e:
                mt.exception_error(e, real_exception=True)
                if 'save_credentials_file' in self.yaml_d:
                    cred_path = self.yaml_d['save_credentials_file']
                    mt.exception_error('{} expired'.format(cred_path), real_exception=False)
                    ts = mt.get_time_stamp()
                    # rename the current just for backup
                    mt.move_file(file_src=cred_path, file_dst='{}.expired_{}'.format(cred_path, ts))
                    # reconnect to get browser log in
                    self.gauth = GoogleAuth(settings_file=yaml_path)
                    self.drive = GoogleDrive(self.gauth)
                    self.__refresh_all_files_list()  # side effect - open a browser to sign in
                else:  # no cred file found
                    raise e

            duplicates_list = []
            for full_path, file_d in self.all_files_dict.items():
                if file_d['id'] == 'Duplicated':
                    duplicates_list.append(full_path)

            if len(duplicates_list) > 0:
                mt.exception_error('Found duplicated files or dirs:', real_exception=False, tabs=tabs)
                for dup in duplicates_list:
                    print(mt.add_color('\t{}{}'.format(tabs * '\t', dup), ops='red'))
                exit(-1)

            self.tabs = tabs
            self.dir_color = dir_color
            self.file_color = file_color
            self.title_color = title_color
        except InvalidConfigError as e:
            print('Ask host to do step HOST.4 in the instructions (create credentials and download them)')
            mt.exception_error('client secrets file no found on path stated in settings.yaml', real_exception=False)
            mt.exception_error(e, real_exception=True)
            exit(-1)
        except RefreshError as e:
            print('delete your credentials.json file and try again')
            mt.exception_error('next run you will be asked to log in and a new credentials.json file will be created',
                               real_exception=False)
            mt.exception_error(e, real_exception=True)
            exit(-1)
        return

    def __str__(self):
        string = '{}{}'.format(self.tabs * '\t', mt.add_color(string='google_handler:', ops=self.title_color))
        string += '\n{}'.format(
            mt.dict_as_table(self.yaml_d, title=os.path.abspath(self.yaml_path), fp=2, ack=False, tabs=self.tabs + 1))
        string += '\n\t{}tabs={}'.format(self.tabs * '\t', self.tabs)
        string += '\n\t{}dir_color={}'.format(self.tabs * '\t', self.dir_color)
        string += '\n\t{}file_color={}'.format(self.tabs * '\t', self.file_color)
        string += '\n\t{}title_color={}'.format(self.tabs * '\t', self.title_color)
        return string

    def __del__(self):
        print('google_handler destroyed')

    def __map_all_files(self, id_str: str = 'root', prefix: str = 'root') -> dict:
        """
        maps all files in google drive for access by name (and not id)
        if a file exist more than once on the same dir - can't use the map. have to use id
        :param id_str: start with root and recursively go over all files id
        :param prefix: the path till this part.
        :return: dict of dicts. each dict has name, id, type. if type is dir - also files list
        """
        file_list = self.drive.ListFile({'q': "'{}' in parents and trashed=false".format(id_str)}).GetList()
        dir_dict = {}
        for file1 in file_list:
            file_name, file_id, file_type = file1['title'], file1['id'], file1['mimeType']
            full_path = '{}/{}'.format(prefix, file_name)

            if full_path in dir_dict:  # maybe found duplicate
                # if file_type == dir_dict[full_path]['type']:  # could be file named 'a' and dir named 'a'
                file_id = 'Duplicated'  # overwriting the id field

            dir_dict[full_path] = {'name': file_name, 'id': file_id, 'type': file_type}
            # print('path: {}, id: {}, type: {}, '.format(full_path, file_id, file_type))
            if file_type.endswith('folder'):
                sub_dict = self.__map_all_files(id_str=file_id, prefix=full_path)
                dir_dict.update(sub_dict)
        return dir_dict

    def __refresh_all_files_list(self) -> None:
        """
        if changes were made - this refresh the all_files_list member
        :return:
        """
        self.all_files_dict = {'root': {'name': 'root', 'id': 'root', 'type': 'application/vnd.google-apps.folder'}}
        all_files_dict = self.__map_all_files(id_str='root', prefix='root')
        self.all_files_dict.update(all_files_dict)
        return

    def __get_file(self, full_path_on_drive: str, tabs: int = 1) -> GoogleDriveFile:
        file = None
        if full_path_on_drive not in self.all_files_dict:
            mt.exception_error(mt.NOT_FOUND.format(full_path_on_drive), real_exception=False, tabs=tabs, depth=3)
        else:
            file = self.drive.CreateFile({'id': self.all_files_dict[full_path_on_drive]['id']})
            if file['mimeType'].endswith('folder'):
                file = None
                mt.exception_error('{} is a dir'.format(full_path_on_drive), real_exception=False, tabs=tabs, depth=3)
        return file

    def __get_dir(self, full_path_on_drive: str, tabs: int = 1) -> GoogleDriveFile:
        _dir = None
        if full_path_on_drive not in self.all_files_dict:
            mt.exception_error(mt.NOT_FOUND.format(full_path_on_drive), real_exception=False, tabs=tabs, depth=3)
        else:
            _dir = self.drive.CreateFile({'id': self.all_files_dict[full_path_on_drive]['id']})
            if not _dir['mimeType'].endswith('folder'):
                _dir = None
                mt.exception_error('{} is a file'.format(full_path_on_drive), real_exception=False, tabs=tabs, depth=3)
        return _dir

    def __create_file_in_dir(self, full_drive_path: str, tabs: int = 1) -> GoogleDriveFile:
        file = None
        dir_path_on_drive, new_file_name = os.path.dirname(full_drive_path), os.path.basename(full_drive_path)

        if dir_path_on_drive not in self.all_files_dict:  # dir must exist
            mt.exception_error(mt.NOT_FOUND.format(dir_path_on_drive), real_exception=False, tabs=tabs, depth=3)
        elif full_drive_path in self.all_files_dict:  # already a file with this name in dir_path_on_drive
            mt.exception_error(mt.EXISTS.format(full_drive_path), real_exception=False, tabs=tabs, depth=3)
        else:
            query = {'parents': [{'id': self.all_files_dict[dir_path_on_drive]['id']}], 'title': new_file_name}
            file = self.drive.CreateFile(metadata=query)
        return file

    def __create_dir_in_dir(self, full_drive_path: str, tabs: int = 1) -> GoogleDriveFile:
        _dir = None
        dir_path_on_drive, dir_name = os.path.dirname(full_drive_path), os.path.basename(full_drive_path)

        if dir_path_on_drive not in self.all_files_dict:  # base dir must exist
            mt.exception_error(mt.NOT_FOUND.format(dir_path_on_drive), real_exception=False, tabs=tabs, depth=3)
        elif full_drive_path in self.all_files_dict:  # already a dir with this name in dir_path_on_drive
            mt.exception_error(mt.EXISTS.format(full_drive_path), real_exception=False, tabs=tabs, depth=3)
        else:
            query = {
                'parents': [{'id': self.all_files_dict[dir_path_on_drive]['id']}],
                'title': dir_name,
                'mimeType': 'application/vnd.google-apps.folder'
            }
            _dir = self.drive.CreateFile(metadata=query)
        return _dir

    def list_all_files(self, starting_dir: str = 'root', tabs: int = 1) -> None:
        """
        prints all files in a directory. if starting_dir is none - prints from root
        :param starting_dir: dir as string e.g. 'root', 'root/test' ...
        :param tabs:
        :return:
        see list_all_files_test()
        """
        title = 'Listing all files from dir {}:'.format(starting_dir)
        msg = mt.add_color(string=title, ops=self.title_color)
        print('{}{}'.format(tabs * '\t', msg))
        if starting_dir not in self.all_files_dict:
            mt.exception_error(mt.NOT_FOUND.format(starting_dir), real_exception=False, tabs=tabs + 1, depth=3)
        else:
            for full_path, file_d in self.all_files_dict.items():
                if full_path.startswith(starting_dir):
                    depth = len(full_path.split('/')) - 1  # number of dirs
                    depth -= len(starting_dir.split('/')) - 1  # normalize to be the zero tabs
                    msg = '{}'.format((depth + tabs) * '\t')
                    if file_d['type'].endswith('folder'):
                        msg += mt.add_color('{}'.format(file_d['name']), ops=self.dir_color)
                    else:
                        msg += mt.add_color('{}'.format(file_d['name']), ops=self.file_color)
                    msg += ' {}'.format(file_d['id'])
                    msg += ' {}'.format(file_d['type'])
                    msg += ' {}'.format(full_path)
                    print(msg)
        return

    def read_file(self, full_path_on_drive: str, tabs: int = 1) -> str:
        """
        if file exists on drive - reads it's content. e.g. json, txt ....
        :param full_path_on_drive:
        :param tabs:
        :return:
        see upload_read_and_delete_test()
        """
        content = None
        file = self.__get_file(full_path_on_drive)
        if file is not None:
            content = file.GetContentString()
            msg = '{}{}'.format(tabs * '\t', mt.LOADED.format(full_path_on_drive))
            print(mt.add_color(msg, ops=mt.SUCCESS_C))
        return content

    def download_file(self, full_path_on_drive: str, local_save_path: str, tabs: int = 1) -> bool:
        """
        if file exists on drive - download to local_save_path
        :param full_path_on_drive:
        :param local_save_path:
        :param tabs:
        :return:
        see upload_read_download_delete_test()
        see upload_download_delete_file_test()
        """
        success = False
        file = self.__get_file(full_path_on_drive)
        if file is not None:
            file.GetContentFile(local_save_path)
            success = True
            file_msg = '{}({})'.format(local_save_path, mt.file_or_folder_size(local_save_path))
            msg = '{}{}'.format(tabs * '\t', mt.DOWNLOADED.format(full_path_on_drive, file_msg))
            print(mt.add_color(msg, ops=mt.SUCCESS_C))
        return success

    def upload_content_to_new_file(self, dst_full_path_on_drive: str, content: str, tabs: int = 1) -> bool:
        """
        :param dst_full_path_on_drive: dir_path_on_drive|new_file_name
        :param content:
        :param tabs:
        :return:
        if dir_path_on_drive exists on drive
            and no file name dst_full_path_on_drive exists:
                upload the content to dst_full_path_on_drive
        see upload_read_download_delete_test()
        """
        success = False
        file = self.__create_file_in_dir(dst_full_path_on_drive)
        if file is not None:
            file.SetContentString(content)
            file.Upload()
            success = True
            msg = '{}{}'.format(tabs * '\t', mt.UPLOADED.format('content', dst_full_path_on_drive))
            print(mt.add_color(msg, ops=mt.SUCCESS_C))
            self.__refresh_all_files_list()
        return success

    def update_file_content(self, file_full_path_on_drive: str, new_content: str, tabs: int = 1) -> bool:
        """
        :param file_full_path_on_drive:
        :param new_content:
        :param tabs:
        :return:
        see update_file_content_test()
        """
        success = False
        file = self.__get_file(file_full_path_on_drive)
        if file is not None:
            file.SetContentString(new_content)
            file.Upload()
            success = True
            msg = '{}{} Changed content'.format(tabs * '\t', file_full_path_on_drive)
            print(mt.add_color(msg, ops=mt.SUCCESS_C))
        return success

    def rename_file(self, file_full_path_on_drive: str, new_name: str, tabs: int = 1) -> bool:
        """
        :param file_full_path_on_drive:
        :param new_name:
        :param tabs:
        :return:
        see rename_file_test()
        """
        success = False
        file = self.__get_file(file_full_path_on_drive)
        if file is not None:
            dir_path_on_drive, new_file_name = os.path.dirname(file_full_path_on_drive), os.path.basename(
                file_full_path_on_drive)
            new_full_path_on_drive = '{}/{}'.format(dir_path_on_drive, new_name)
            if new_full_path_on_drive not in self.all_files_dict:
                file['title'] = new_name
                file.Upload()
                success = True
                msg = '{}{} Renamed to {}'.format(tabs * '\t', file_full_path_on_drive, new_full_path_on_drive)
                print(mt.add_color(msg, ops=mt.SUCCESS_C))
                self.__refresh_all_files_list()
            else:
                mt.exception_error(mt.EXISTS.format(new_full_path_on_drive), real_exception=False, tabs=tabs, depth=2)
        return success

    def upload_file(self, dst_full_path_on_drive: str, local_file_path: str, tabs: int = 1) -> bool:
        """
        :param dst_full_path_on_drive: dir_path_on_drive|new_file_name
        :param local_file_path:
        :param tabs:
        :return:
        if dir_path_on_drive exists on drive
            and no file name full_drive_path exists:
                upload the local_file_path to dst_full_path_on_drive
        see upload_delete_image_test()
        """
        success = False
        file = self.__create_file_in_dir(dst_full_path_on_drive)
        if file is not None:
            file.SetContentFile(local_file_path)
            file.Upload()
            success = True
            file_msg = '{}({})'.format(local_file_path, mt.file_or_folder_size(local_file_path))
            msg = '{}{}'.format(tabs * '\t', mt.UPLOADED.format(file_msg, dst_full_path_on_drive))
            print(mt.add_color(msg, ops=mt.SUCCESS_C))
            self.__refresh_all_files_list()
        return success

    def delete_file(self, full_path_on_drive: str, tabs: int = 1) -> bool:
        """
        deletes file if exists
        :param full_path_on_drive:
        :param tabs:
        :return:
        Future - add option to send to trash and restore from trash
        file.Trash()  # Move file to trash.
        file.UnTrash()  # Move file out of trash.
        see upload_delete_image_test()
        """
        success = False
        file = self.__get_file(full_path_on_drive)
        if file is not None:
            file.Delete()
            success = True
            msg = '{}{}'.format(tabs * '\t', mt.DELETED.format(full_path_on_drive))
            print(mt.add_color(msg, ops=mt.SUCCESS_C))
            self.__refresh_all_files_list()
        return success

    def create_dir(self, dst_full_path_on_drive: str, tabs: int = 1) -> bool:
        """
        :param dst_full_path_on_drive: dir_path_on_drive|new_dir_name
        :param tabs:
        :return:
        if dir_path_on_drive exists on drive
            and no dir name dst_full_path_on_drive exists:
                create a new dir named new_dir_name
        see create_and_delete_empty_dir_test()
        """
        success = False
        _dir = self.__create_dir_in_dir(dst_full_path_on_drive)
        if _dir is not None:
            _dir.Upload()
            success = True
            msg = '{}{}'.format(tabs * '\t', mt.CREATED.format(dst_full_path_on_drive))
            print(mt.add_color(msg, ops=mt.SUCCESS_C))
            self.__refresh_all_files_list()
        return success

    def delete_empty_dir(self, full_path_on_drive, tabs: int = 1) -> bool:
        """
        :param full_path_on_drive:
        :param tabs:
        :return:
        see create_and_delete_empty_dir_test()
        """
        success = False
        _dir = self.__get_dir(full_path_on_drive)
        if _dir is not None:
            is_empty = True
            for file_path in self.all_files_dict.keys():
                if file_path.startswith(full_path_on_drive) and file_path != full_path_on_drive:
                    is_empty = False
            if is_empty:
                _dir.Delete()
                success = True
                msg = '{}{}'.format(tabs * '\t', mt.DELETED.format(full_path_on_drive))
                print(mt.add_color(msg, ops=mt.SUCCESS_C))
                self.__refresh_all_files_list()
            else:
                mt.exception_error('{} is not empty. use delete_dir_with_files()'.format(full_path_on_drive),
                                   real_exception=False, tabs=tabs, depth=2)
        return success

    def delete_dir_with_files(self, full_path_on_drive, tabs: int = 1) -> bool:
        """
        :param full_path_on_drive:
        :param tabs:
        :return:
        see create_and_delete_dir_with_files_test()
        """
        success = False
        _dir = self.__get_dir(full_path_on_drive)
        if _dir is not None:
            files_c = 0
            for file_path in self.all_files_dict.keys():
                if file_path.startswith(full_path_on_drive) and file_path != full_path_on_drive:
                    files_c += 1

            _dir.Delete()
            success = True
            _dir_msg = '{}({} files)'.format(full_path_on_drive, files_c)
            msg = '{}{}'.format(tabs * '\t', mt.DELETED.format(_dir_msg))
            print(mt.add_color(msg, ops=mt.SUCCESS_C))
            self.__refresh_all_files_list()
        return success

    def download_dir(self, full_path_on_drive: str, local_dir_name: str, tabs: int = 1) -> bool:
        """
        :param full_path_on_drive:
        :param local_dir_name:
        :param tabs:
        :return:
        see download_dir_test()
        """
        success = False
        if os.path.exists(local_dir_name):
            mt.exception_error(mt.EXISTS.format(local_dir_name), tabs=tabs, depth=2)
        else:
            _dir = self.__get_dir(full_path_on_drive)
            if _dir is not None:
                print('{}Creating local dirs and downloading files:'.format(tabs * '\t'))
                mt.create_dir(dir_path=local_dir_name, tabs=tabs + 1)
                # create all dirs locally
                for full_path, file_d in self.all_files_dict.items():
                    if full_path.startswith(full_path_on_drive) and full_path != full_path_on_drive:
                        if file_d['type'].endswith('folder'):
                            rel_path = full_path[len(full_path_on_drive) + 1:]
                            mt.create_dir('{}/{}'.format(local_dir_name, rel_path), tabs=tabs + 1)

                # download all file to existing dir structure
                for full_path, file_d in self.all_files_dict.items():
                    if full_path.startswith(full_path_on_drive) and full_path != full_path_on_drive:
                        if not file_d['type'].endswith('folder'):
                            rel_path = full_path[len(full_path_on_drive) + 1:]
                            save_path = '{}/{}'.format(local_dir_name, rel_path)
                            self.download_file(full_path_on_drive=full_path, local_save_path=save_path, tabs=tabs + 1)

                file_msg = '{}({})'.format(local_dir_name, mt.file_or_folder_size(local_dir_name))
                msg = '{}{}'.format(tabs * '\t', mt.DOWNLOADED.format(full_path_on_drive, file_msg))
                print(mt.add_color(msg, ops=mt.SUCCESS_C))
                success = True
        return success

    def upload_dir(self, dst_full_path_on_drive: str, local_dir_name: str, tabs: int = 1) -> bool:
        """
        :param dst_full_path_on_drive:
        :param local_dir_name:
        :param tabs:
        :return:
        see upload_dir_test()
        """
        success = False
        if not os.path.exists(local_dir_name):
            mt.exception_error(mt.NOT_FOUND.format(local_dir_name), tabs=tabs, depth=2)
        else:
            print('{}Creating dirs and uploading files:'.format(tabs * '\t'))
            success_dir = self.create_dir(dst_full_path_on_drive, tabs=tabs + 1)
            if success_dir:
                # create all dirs on drive
                empty_dirs = []
                for _dir_path, dir_names, file_names in os.walk(local_dir_name):
                    for d in dir_names:
                        empty_dirs.append(d)
                for ed in empty_dirs:
                    self.create_dir('{}/{}'.format(dst_full_path_on_drive, ed), tabs=tabs + 1)

                # upload all files
                files_full_path = {}
                for _dir_path, dir_names, file_names in os.walk(local_dir_name):
                    for f in file_names:
                        # local full path
                        fp_local = '{}/{}'.format(os.path.abspath(_dir_path).replace('\\', '/'), f)

                        # drive relative path
                        fp_relative_dir = os.path.abspath(_dir_path)[len(os.path.abspath(local_dir_name)) + 1:]
                        if fp_relative_dir == '':
                            fp_drive_rel = f
                        else:
                            fp_drive_rel = '{}/{}'.format(fp_relative_dir.replace('\\', '/'), f)
                        fp_on_drive = '{}/{}'.format(dst_full_path_on_drive, fp_drive_rel)  # add 'root/.../...'

                        files_full_path[fp_local] = fp_on_drive
                for fp_local_rel, fp_on_drive in files_full_path.items():
                    self.upload_file(dst_full_path_on_drive=fp_on_drive, local_file_path=fp_local_rel, tabs=tabs + 1)

                file_msg = '{}({})'.format(local_dir_name, mt.file_or_folder_size(local_dir_name))
                msg = '{}{}'.format(tabs * '\t', mt.UPLOADED.format(file_msg, dst_full_path_on_drive))
                print(mt.add_color(msg, ops=mt.SUCCESS_C))
                success = True
                self.__refresh_all_files_list()

        return success



from tflite_runtime.interpreter import Interpreter
import numpy as np
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.open_cv import open_cv_tools as cvt
from wizzi_utils.models import models_configs as cfg
from wizzi_utils.models.base_models import PdBaseModel
# noinspection PyPackageRequirements
import cv2


class TflPdModel(PdBaseModel):
    def __init__(
            self,
            save_load_dir: str,
            model_name: str,
            threshold: float = None,
            allowed_joint_names: list = None,
            min_p_valid_joints: float = 0.3,
            check_type: bool = True
    ):
        """
        :param save_load_dir: where the model is saved (or will be if not exists)
        :param model_name: valid name in MODEL_CONF.keys()
        :param threshold: only detection above this threshold will be pass first filter
        :param allowed_joint_names: joint_names to track from the joint_names of the model
        :param min_p_valid_joints:
            let x=min_p_valid_joints*len(allowed_joint_names)
            to keep a pose: we need max(x,2) joints found
        :param check_type:
        example:
        see:
        """
        super().__init__(save_load_dir=save_load_dir, model_name=model_name, allowed_joint_names=allowed_joint_names,
                         min_p_valid_joints=min_p_valid_joints)
        if check_type and not self.model_type_valid(self.model_cfg['model_type'], cfg.ModelType.PdTflNormal.value):
            exit(-1)
        if threshold is not None:
            self.model_cfg['threshold'] = threshold
        self.pairs_indices = self.model_cfg['pairs_indices']

        model_fp = "{}/{}.tflite".format(self.local_path, self.model_name)
        self._download_if_needed(local_path=model_fp, url_dict=self.model_cfg['tflite'])
        self.model_size = mt.file_or_folder_size(model_fp)
        # print('Loading {}(size {}, {} classes)'.format(self.model_name, self.model_size, len(self.labels)))
        self.interpreter = Interpreter(model_path=model_fp, num_threads=4)
        # allocate input output placeholders
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        _, input_height, input_width, _ = self.interpreter.get_input_details()[0]['shape']
        self.model_cfg['in_dims'] = (input_width, input_height)
        self.model_cfg['input_type'] = self.interpreter.get_input_details()[0]['dtype']
        quantization = self.interpreter.get_input_details()[0]['quantization']
        # need to normalize if model doesn't do quantization (then quantization == (0.0, 0))
        # else e.g. quantization == (0.0078125, 128) - no need to normalize
        self.model_cfg['normalize_RGB'] = True if quantization == (0.0, 0) else False

        # # you can print this to get more details on the model
        # mt.dict_as_table(self.interpreter.get_input_details()[0], title='input')
        # mt.dict_as_table(self.interpreter.get_output_details()[0], title='output')
        # mt.dict_as_table(self.interpreter.get_tensor_details()[0], title='tensor')
        return

    def to_string(self, tabs: int = 1) -> str:
        tabs_s = tabs * '\t'
        string = '{}{}'.format(tabs_s, mt.add_color(string='TflPdModel:', ops='underlined'))
        string += '\n\t{}name= {} (size {})'.format(tabs_s, self.model_name, self.model_size)
        string += '\n\t{}local_path= {}'.format(tabs_s, self.local_path)
        string += '\n\t{}{}'.format(tabs_s, mt.to_str(self.allowed_joint_names, 'allowed_joint_names'))
        string += '\n{}'.format(mt.dict_as_table(self.model_cfg, title='cfg', fp=6, ack=False, tabs=tabs + 1))
        return string

    def prepare_input(self, cv_img: np.array) -> np.array:
        """
        :param cv_img:
        resize and change dtype to predefined params
        :return:
        """
        img_RGB = cvt.BGR_img_to_RGB(cv_img)

        if self.model_cfg['normalize_RGB']:
            center = 127.5
            img_RGB = (img_RGB / center) - 1  # normalize image
        img = cv2.resize(img_RGB, self.model_cfg['in_dims'])  # size of this model input
        img_processed = np.expand_dims(img, axis=0).astype(self.model_cfg['input_type'])  # a,b,c -> 1,a,b,c
        return img_processed

    def run_network(self, img_preprocessed: np.array) -> None:
        self.interpreter.set_tensor(self.input_details[0]['index'], img_preprocessed)  # set input tensor
        self.interpreter.invoke()  # run
        return

    def extract_results(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: cv image
        :param fp: float precision on the score percentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts:
        dict is a detection of pose
            each has items:
                'joint_names_list':  - all joint names like in the config
                'joint_ids_list': - all joint ids like in the config
                'joint_x_y_list': - if joint found: it's x,y values
                'joint_z_list': - if joint found: it's z value
                'score_percentages_list': - if joint found: it's confidence in 0-100%,
        """
        # get results
        depth = 5  # each points has x,y,z,visibility,presence
        # full pose -> 195/5=39. but i think there are only 33
        outputs = self.interpreter.get_tensor(self.output_details[0]['index'])[0]  # points numpy

        if ack:
            title_suffix = '' if img_title is None else '{} '.format(img_title)
            title = '{} detection on image {}{}:'.format(self.model_name, title_suffix, cv_img.shape)
            print('{}{}'.format(tabs * '\t', title))
            print('{}Meta_data:'.format(tabs * '\t'))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(outputs, 'outputs')))
            print('{}Detections:'.format(tabs * '\t'))

        detections = []  # each detection is a set of joint
        img_h, img_w = cv_img.shape[0], cv_img.shape[1]
        net_w, net_h = self.model_cfg['in_dims']
        for pose_id in range(1):  # currently supports 1 pose
            detection_d = {
                'joint_names_list': [],
                'joint_ids_list': [],
                'joint_x_y_list': [],
                'joint_z_list': [],
                'score_percentages_list': [],
                'bbox': {}
            }
            valid_joints_found = 0  # max is len(self.allowed_joint_names)
            for joint_id, joint_name in self.joint_names.items():
                detection_d['joint_names_list'].append(joint_name)
                detection_d['joint_ids_list'].append(joint_id)
                x = int(outputs[joint_id * depth + 0] * img_w / net_w)
                y = int(outputs[joint_id * depth + 1] * img_h / net_h)
                z = int(outputs[joint_id * depth + 2])

                visibility = outputs[joint_id * depth + 3]
                visibility = 1 / (1 + np.exp(visibility))  # reverse sigmoid
                presence = outputs[joint_id * depth + 4]
                presence = 1 / (1 + np.exp(presence))  # reverse sigmoid
                score_frac = 1 - max(visibility, presence)  # change from err to acc: acc = 1-err
                score_percentage = round(score_frac * 100, fp)

                if self.model_cfg['threshold'] <= score_frac <= 1.0 and joint_name in self.allowed_joint_names:
                    detection_d['joint_x_y_list'].append([x, y])
                    detection_d['joint_z_list'].append(z)
                    detection_d['score_percentages_list'].append(score_percentage)
                    valid_joints_found += 1
                else:
                    detection_d['joint_x_y_list'].append(self.NOT_FOUND_PAIR)
                    detection_d['joint_z_list'].append(self.NOT_FOUND_VALUE)
                    detection_d['score_percentages_list'].append(self.NOT_FOUND_VALUE)

                if ack:
                    d_msg = '{}\tpose {}: {}({}): xy={}, z={}, score=({}%)'
                    print(d_msg.format(tabs * '\t', pose_id, joint_name, joint_id, detection_d['joint_x_y_list'][-1],
                                       detection_d['joint_z_list'][-1], detection_d['score_percentages_list'][-1]))
            min_joints_found_ok = valid_joints_found >= self.min_p_valid_joints * len(self.allowed_joint_names)
            pose_valid = min_joints_found_ok and valid_joints_found >= 2
            if pose_valid:  # minimum 2 regardless to self.min_p_valid_joints(which could be zero)
                detection_d['bbox'] = PdBaseModel.get_bbox_dict(detection_d['joint_x_y_list'])
                if detection_d['bbox'] is not None:
                    if ack:
                        print('{}\tbbox:{}'.format(tabs * '\t', detection_d['bbox']))
                    detections.append(detection_d)
                else:
                    if ack:
                        print('{}\tPose {} was dumped. bbox is None'.format(tabs * '\t', pose_id))
            else:
                if ack:
                    min_needed = max(self.min_p_valid_joints * len(self.allowed_joint_names), 2)
                    msg = '{}\tPose {} was dumped. It had valid_joints_found={} (min needed {})'
                    print(msg.format(tabs * '\t', pose_id, valid_joints_found, min_needed))
        return detections

    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score precentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """

        img_preprocessed = self.prepare_input(cv_img)
        self.run_network(img_preprocessed)
        detections = self.extract_results(
            cv_img=cv_img,
            fp=fp,
            ack=ack,
            tabs=tabs,
            img_title=img_title
        )
        return detections


class TflPdModelPoseNet(TflPdModel):
    def __init__(
            self,
            save_load_dir: str,
            model_name: str,
            threshold: float = None,
            allowed_joint_names: list = None,
            min_p_valid_joints: float = 0.3,
            check_type: bool = True
    ):
        super().__init__(save_load_dir=save_load_dir, model_name=model_name, threshold=threshold,
                         allowed_joint_names=allowed_joint_names, min_p_valid_joints=min_p_valid_joints,
                         check_type=False)
        if check_type and not self.model_type_valid(self.model_cfg['model_type'], cfg.ModelType.PdTflPoseNet.value):
            exit(-1)
        return

    @staticmethod
    def mod(a: np.array, b: int) -> np.array:
        """ find a % b """
        floored = np.floor_divide(a, b)
        return np.subtract(a, np.multiply(floored, b))

    @staticmethod
    def sigmoid(x: np.array) -> np.array:
        """ apply sigmoid activation to numpy array """
        return 1 / (1 + np.exp(-x))

    @staticmethod
    def sigmoid_and_argmax2d(heatmap: np.array, threshold: float) -> np.array:
        """
        :param heatmap: 9x9x17 (17 joints)
        :param threshold:
        :return: y,x coordinates 17x2
        """
        # v1 is 9x9x17 heatmap
        # v1 = interpreter.get_tensor(output_details[0]['index'])[0]
        height, width, depth = heatmap.shape
        reshaped = np.reshape(heatmap, [height * width, depth])
        reshaped = TflPdModelPoseNet.sigmoid(reshaped)
        # apply threshold
        reshaped = (reshaped > threshold) * reshaped
        coords = np.argmax(reshaped, axis=0)
        yCoords = np.round(np.expand_dims(np.divide(coords, width), 1))
        xCoords = np.expand_dims(TflPdModelPoseNet.mod(coords, width), 1)
        ret = np.concatenate([yCoords, xCoords], 1)
        return ret

    @staticmethod
    def get_offsets(offsets: np.array, coords: np.array, num_key_points: int = 17) -> np.array:
        """
        :param offsets: 9x9x34 - probably yx heatmap per joint(17*2)
        :param coords: 17x2
        :param num_key_points: number of joints
        :return: get offset vectors from all coordinates
        """
        # offsets = interpreter.get_tensor(output_details[1]['index'])[0]
        offset_vectors = []
        for i, (heatmap_y, heatmap_x) in enumerate(coords):
            # print(i, y, x)
            heatmap_y = int(heatmap_y)
            heatmap_x = int(heatmap_x)
            # print(heatmap_y, heatmap_x)
            # make sure indices aren't out of range
            heatmap_y = min(8, heatmap_y)
            heatmap_x = min(8, heatmap_x)
            y_off = offsets[heatmap_y, heatmap_x, i]
            x_off = offsets[heatmap_y, heatmap_x, i + num_key_points]
            # ov = get_offset_point(heatmap_y, heatmap_x, offsets, i, num_key_points)
            offset_vectors.append([y_off, x_off])
        offset_vectors = np.array(offset_vectors)
        return offset_vectors

    def extract_results(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: cv image
        :param fp: float precision on the score percentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts:
        dict is a detection of pose
            each has items:
                'joint_names_list':  - all joint names like in the config
                'joint_ids_list': - all joint ids like in the config
                'joint_x_y_list': - if joint found: it's x,y values
                'joint_z_list': - if joint found: it's z value
                'score_percentages_list': - if joint found: it's confidence in 0-100%,
        """
        outputs = self.interpreter.get_tensor(self.output_details[0]['index'])[0]
        offsets = self.interpreter.get_tensor(self.output_details[1]['index'])[0]

        if ack:
            title_suffix = '' if img_title is None else '{} '.format(img_title)
            title = '{} detection on image {}{}:'.format(self.model_name, title_suffix, cv_img.shape)
            print('{}{}'.format(tabs * '\t', title))
            print('{}Meta_data:'.format(tabs * '\t'))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(outputs, 'outputs')))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(offsets, 'offsets')))
            print('{}Detections:'.format(tabs * '\t'))

        # get y,x positions from heat map
        yx = TflPdModelPoseNet.sigmoid_and_argmax2d(outputs, threshold=self.model_cfg['threshold'])
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(yx, 'yx')))

        # points below threshold (value is [0, 0])
        drop_pts = list(np.unique(np.where(yx == 0)[0]))
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(drop_pts, 'drop_pts')))

        # get offsets from positions
        offset_vectors = TflPdModelPoseNet.get_offsets(offsets, yx)
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(offset_vectors, 'offset_vectors')))

        # use stride to get coordinates in image coordinates
        output_stride = 32
        yx_values = yx * output_stride + offset_vectors
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(yx_values, 'yx_values')))

        for bad_joint_ind in drop_pts:
            yx_values[bad_joint_ind] = self.NOT_FOUND_PAIR
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(yx_values, 'yx_values')))

        detections = []  # each detection is a set of joint
        img_h, img_w = cv_img.shape[0], cv_img.shape[1]
        net_w, net_h = self.model_cfg['in_dims']
        for pose_id in range(1):  # currently supports 1 pose
            detection_d = {
                'joint_names_list': [],
                'joint_ids_list': [],
                'joint_x_y_list': [],
                'score_percentages_list': [],
                'bbox': {}
            }
            valid_joints_found = 0  # max is len(self.allowed_joint_names)
            for joint_id, joint_name in self.joint_names.items():
                detection_d['joint_names_list'].append(joint_name)
                detection_d['joint_ids_list'].append(joint_id)

                if yx_values[joint_id].tolist() == self.NOT_FOUND_PAIR:
                    score_frac = 0
                    x, y, score_percentage = None, None, None
                else:
                    y = int(yx_values[joint_id][0] * img_h / net_h)
                    x = int(yx_values[joint_id][1] * img_w / net_w)
                    score_frac = 0.99  # for now till i will extract the confidence from the detection
                    # score_percentage = round(score_frac * 100, fp)
                    score_percentage = 'TODO ???'

                if self.model_cfg['threshold'] <= score_frac <= 1.0 and joint_name in self.allowed_joint_names:
                    detection_d['joint_x_y_list'].append([x, y])
                    detection_d['score_percentages_list'].append(score_percentage)
                    valid_joints_found += 1
                else:
                    detection_d['joint_x_y_list'].append(self.NOT_FOUND_PAIR)
                    detection_d['score_percentages_list'].append(self.NOT_FOUND_VALUE)

                if ack:
                    d_msg = '{}\tpose {}: {}({}): xy={}, score=({}%)'
                    print(d_msg.format(tabs * '\t', pose_id, joint_name, joint_id, detection_d['joint_x_y_list'][-1],
                                       detection_d['score_percentages_list'][-1]))
            min_joints_found_ok = valid_joints_found >= self.min_p_valid_joints * len(self.allowed_joint_names)
            pose_valid = min_joints_found_ok and valid_joints_found >= 2
            if pose_valid:  # minimum 2 regardless to self.min_p_valid_joints(which could be zero)
                detection_d['bbox'] = PdBaseModel.get_bbox_dict(detection_d['joint_x_y_list'])
                if detection_d['bbox'] is not None:
                    if ack:
                        print('{}\tbbox:{}'.format(tabs * '\t', detection_d['bbox']))
                    detections.append(detection_d)
                else:
                    if ack:
                        print('{}\tPose {} was dumped. bbox is None'.format(tabs * '\t', pose_id))
            else:
                if ack:
                    min_needed = max(self.min_p_valid_joints * len(self.allowed_joint_names), 2)
                    msg = '{}\tPose {} was dumped. It had valid_joints_found={} (min needed {})'
                    print(msg.format(tabs * '\t', pose_id, valid_joints_found, min_needed))
        return detections



from wizzi_utils.misc import misc_tools as mt


def test_cv2_object_detection_models():
    try:
        from wizzi_utils.models.test import test_cv2_od

        test_cv2_od.test_all()
    except (ModuleNotFoundError, AttributeError) as e:
        mt.exception_error(e)
        print('\tIt\'s possible you dont have open cv')
    return


def test_tflite_object_detection_models():
    try:
        from wizzi_utils.models.test import test_tflite_od

        test_tflite_od.test_all()
    except (ModuleNotFoundError, AttributeError) as e:
        mt.exception_error(e)
        print('\tIt\'s possible you dont have tflite')
    return


def test_cv2_pose_detection_models():
    try:
        from wizzi_utils.models.test import test_cv2_pd

        test_cv2_pd.test_all()
    except (ModuleNotFoundError, AttributeError) as e:
        mt.exception_error(e)
        print('\tIt\'s possible you dont have open cv')
    return


def test_tflite_pose_detection_models():
    try:
        from wizzi_utils.models.test import test_tflite_pd

        test_tflite_pd.test_all()
    except (ModuleNotFoundError, AttributeError) as e:
        mt.exception_error(e)
        print('\tIt\'s possible you dont have tflite')
    return


def test_cv2_tracking_models():
    try:
        from wizzi_utils.models.test import test_cv2_tracking

        test_cv2_tracking.test_all()
    except (ModuleNotFoundError, AttributeError) as e:
        mt.exception_error(e)
        print('\tIt\'s possible you dont have open cv contrib')
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    test_cv2_object_detection_models()
    test_tflite_object_detection_models()
    test_cv2_pose_detection_models()
    test_tflite_pose_detection_models()
    test_cv2_tracking_models()
    print('{}'.format('-' * 20))
    return



from wizzi_utils.coreset import coreset_tools as cot
from wizzi_utils.misc import misc_tools as mt
import numpy as np


def select_coreset_test():
    mt.get_function_name(ack=True, tabs=0)
    A = np.array([[1, 1], [2, 1], [3, 7], [4, 4], [5, 1]])
    SP = np.ones(shape=A.shape[0]) / A.shape[0]  # uniform probability
    C, CW = cot.select_coreset(A=A, SP=SP, c_size=2, with_reps=True)
    print(mt.to_str(var=A, title='\tA'))
    print(mt.to_str(var=C, title='\tC'))
    print(mt.to_str(var=CW, title='\tCW'))
    return


def select_coreset_frequency_test():
    mt.get_function_name(ack=True, tabs=0)
    A = np.array([[1, 1], [2, 1], [3, 7], [4, 4], [5, 1]])
    SP = np.ones(shape=A.shape[0]) / A.shape[0]  # uniform probability
    C, CW = cot.select_coreset_frequency(P=A, SP=SP, c_size=4)
    print(mt.to_str(var=A, title='\tA'))
    print(mt.to_str(var=C, title='\tC'))
    print(mt.to_str(var=CW, title='\tCW'))
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    select_coreset_test()
    select_coreset_frequency_test()
    print('{}'.format('-' * 20))
    return



# noinspection PyPackageRequirements
import cv2
from enum import Enum
import numpy as np
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.open_cv import open_cv_tools as cvt
from wizzi_utils.pyplot import pyplot_tools as pyplt


# noinspection PyUnresolvedReferences
class Trackers(Enum):
    # fps relate to scooter on this vid - not clear if running on cpu\gpu
    CSRT = cv2.legacy.TrackerCSRT_create  # 48.05 FPS
    KCF = cv2.legacy.TrackerKCF_create  # 170.63 FPS
    BOOSTING = cv2.legacy.TrackerBoosting_create  # 55 FPS
    MIL = cv2.legacy.TrackerMIL_create  # 14.79 FPS
    TLD = cv2.legacy.TrackerTLD_create  # 30 FPS
    MF = cv2.legacy.TrackerMedianFlow_create  # 800 FPS
    MOSSE = cv2.legacy.TrackerMOSSE_create  # 3000 FPS
    GOTURN = cv2.TrackerGOTURN_create  # error


class Cv2Tracker:
    def __init__(self, model_obj: Trackers):
        """
        :param model_obj: from class Trackes(Enum)
        """
        if model_obj not in Trackers:
            exit(-1)
        self.tracker_type = model_obj.name
        self.tracker_ctor = model_obj.value
        self.trackers = []
        return

    def to_string(self, tabs: int = 1) -> str:
        tabs_s = tabs * '\t'
        string = '{}{}'.format(tabs_s, mt.add_color(string='Cv2Tracker:', ops='underlined'))
        string += '\n\t{}tracker_type={}'.format(tabs_s, self.tracker_type)
        string += '\n\t{}supports multi objects tracking'.format(tabs_s)
        return string

    def add_obj_bbox_manual(self, cv_img: np.array, xywh: tuple, frame_idx: int, name: str = None, ack: bool = False):
        """
        :param cv_img:
        :param xywh:
        :param ack:
        :param frame_idx:
        :param name:
        :return:
        """
        if not xywh == (0, 0, 0, 0):  # clicked space\enter after selecting a bbox
            xywh = self.tuple_to_int(xywh)
            tracker = self.tracker_ctor()
            tracker.init(cv_img, xywh)
            tracker_idx = len(self.trackers)
            tracker_dict = {
                'active': True,  # will be set to off when user decides no tracking needed (e.g. object left the frame)
                'type': self.tracker_type,  # kcf, csrt ... not used for now
                'idx': tracker_idx,  # idx in self.tracker
                # used for draw labels
                'name': '{}({})'.format(name, tracker_idx) if name is not None else str(tracker_idx),
                'obj': tracker,  # the actual tracker
                'start_info': {  # debug and analysis for now
                    'post_frame': frame_idx,
                    'init_xywh': xywh,
                    'image_dims': cv_img.shape
                },
                'bboxes': {},  # will contain all bboxes of object found over iters - post run calculation
                'tracks': []  # will contain all center of mass points from previous iterations - draw tracks
            }
            self.trackers.append(tracker_dict)

            if ack:
                tl, br = self.convert_xywh_to_tl_and_br(xywh=xywh)
                msg = '\tpost frame {}: tracker {} created - tracking object on tl={}, br={} (xywh={})'
                print(msg.format(frame_idx, len(self.trackers), tl, br, xywh))
        return

    def add_obj_bbox_roi(self, cv_title: str, cv_img: np.array, frame_idx: int, ack: bool = False) -> None:
        """
        :param cv_title:
        :param cv_img:
        :param ack:
        :param frame_idx:
        :return:
        give an option to select a bbox and add to the trackers
        since we want to do it on the already open cv window, we pass title and we can't reshape the cv_img.
        the bbox that we will get from ROI will be in the cv_img dims. need to convert to self.in_dims
        """
        xywh = cv2.selectROI(cv_title, cv_img, fromCenter=False, showCrosshair=True)
        self.add_obj_bbox_manual(cv_img=cv_img, xywh=xywh, frame_idx=frame_idx, ack=ack)
        return

    def track_frame(
            self,
            cv_img: np.array,
            frame_idx: int,
            ack: bool = False,
            tabs: int = 1,
    ) -> list:
        detections = []

        if ack:
            print('{}Frame {} - image shape = {}:'.format(tabs * '\t', frame_idx, cv_img.shape))

        for idx, tracker_dict in enumerate(self.trackers):
            if tracker_dict['active']:
                (success, xywh) = tracker_dict['obj'].update(cv_img)
                if success:
                    xywh = self.tuple_to_int(xywh)
                    (x0, y0), (x1, y1) = self.convert_xywh_to_tl_and_br(xywh=xywh)
                    center_x = (x0 + x1) / 2
                    center_y = (y0 + y1) / 2
                    center = (center_x, center_y)
                    center = self.tuple_to_int(center)
                    tracker_dict['tracks'].append(center)
                    if ack:
                        msg = '{}\tTrackerID={}: success. object on tl={}, br={} (center={}, xywh={})'
                        print(msg.format(tabs * '\t', idx, (x0, y0), (x1, y1), center, xywh))
                else:
                    x0, y0, x1, y1 = -1, -1, -1, -1
                    if ack:
                        print('{}\tTrackerID={}: object not found'.format(tabs * '\t', idx))
                    # TODO go over last 'x' detections - if all false - remove from active tracker

                detection_d = {
                    'tracker_idx': idx,
                    'success': success,
                    'bbox': {
                        #  pt1 = (x0, y0)  # obj frame top left corner
                        #  pt2 = (x1, y1)  # obj frame bottom right corner
                        'x0': x0,
                        'y0': y0,
                        'x1': x1,
                        'y1': y1,
                    },
                }

                detections.append(detection_d)
                tracker_dict['bboxes'][frame_idx] = detection_d  # save for future calculation. e.g. calculate direction
        return detections

    @staticmethod
    def convert_xywh_to_tl_and_br(xywh: tuple) -> (tuple, tuple):
        x0, y0, w, h = xywh
        x1, y1 = x0 + w, y0 + h
        return (x0, y0), (x1, y1)

    @staticmethod
    def convert_tl_and_br_to_xywh(tl: tuple, br: tuple) -> (tuple, tuple):
        x0, y0 = tl
        x1, y1 = br
        xywh = (x0, y0, x1 - x0, y1 - y0)
        return xywh

    @staticmethod
    def change_dims(xywh: tuple, old_dims: tuple, new_dims: tuple) -> tuple:
        # change scale to in_dims
        (x, y, w, h) = xywh
        new_w, new_h = new_dims
        old_w, old_h = old_dims
        x *= new_w / old_w
        w *= new_w / old_w
        y *= new_h / old_h
        h *= new_h / old_h
        return x, y, w, h

    @staticmethod
    def tuple_to_int(t: tuple) -> tuple:
        new_t = tuple([int(item) for item in t])
        return new_t

    def get_tracker_dict(self, tracker_id: int) -> dict:
        tr_dict = None
        if len(self.trackers) >= tracker_id + 1:
            tr_dict = self.trackers[tracker_id]
        return tr_dict

    def draw_detections(
            self,
            detections: list,
            cv_img: np.array,
            colors_d: dict = None,
            draw_labels: bool = False,
            tracks_dict: dict = None,
            header_tl: dict = None,
            header_bl: dict = None,
            header_tr: dict = None,
            header_br: dict = None,
    ) -> None:
        """
        :param detections: output of self.classify_cv_img() on od models
        :param colors_d: colors for the bounding boxes in str form:
            if tracker id exist - color the bbox of that tracker in the given color
            e.g. colors_d={
                    'default_bbox_color': 'black',
                    'bbox_color_0': 'blue',  # custom color for tracker 0 (if active and made detection)
                    'bbox_color_1': 'red',  # custom color for tracker 1 (if active and made detection)
                },
        :param draw_labels: if True draw tracker "name" given on creation
            "name" by default is tracker idx
        :param tracks_dict: if to draw the "tracks" of the object
            one track:= the center of the bbox from a previous iteration
            if None - no tracks
            if not None:
            e.g.
            tracks_dict = {
                    'default_tracks_color': 'blue',  # circle color
                    'len_tracks': 100,  # how much bbox centers will be drawn
                    'radius': 1,  # circle radius (size)
                    'thickness': -1 # circle thickness (-1 is filled)
                    'tracks_color_1': 'red' # custom color for tracker 1 (if active and made detection)
                }
        :param cv_img: the same that was given input to self.classify_cv_img()
        :param header_tl: dict with args for cvt.add_header on top_left corner
        :param header_bl: dict with args for cvt.add_header on bottom_left corner
        :param header_tr: dict with args for cvt.add_header on top_right corner
        :param header_br: dict with args for cvt.add_header on bottom_right corner
        header e.g.
        header_tl={'text': 'wizzi_utils', 'x_offset': 0, 'text_color': 'aqua', 'with_rect': True, 'bg_color': 'black',
               'bg_font_scale': 1}
        :return:
        """
        if colors_d is None:
            colors_d = {
                'default_bbox_color': 'lime'
            }
        for detection in detections:
            success = detection['success']
            if success:
                tracker_idx = detection['tracker_idx']
                tracker_obj = self.get_tracker_dict(tracker_id=tracker_idx)
                x0 = detection['bbox']['x0']
                y0 = detection['bbox']['y0']
                x1 = detection['bbox']['x1']
                y1 = detection['bbox']['y1']

                # DRAW BBOX
                k = 'bbox_color_{}'.format(tracker_idx)  # check custom color for this tracker
                bbox_color_bgr = pyplt.get_BGR_color(colors_d[k]) if k in colors_d else pyplt.get_BGR_color(
                    colors_d['default_bbox_color'])
                cv2.rectangle(img=cv_img, pt1=(x0, y0), pt2=(x1, y1), color=bbox_color_bgr, thickness=2)

                if draw_labels:
                    label = tracker_obj['name']
                    cvt.add_text(cv_img, header=label, pos=(x0, y1), text_color='w', with_rect=True,
                                 bg_color='black', bg_font_scale=2)

                if tracks_dict is not None:
                    k = 'tracks_color_{}'.format(tracker_idx)  # check custom color for this tracker
                    tracks_color_bgr = pyplt.get_BGR_color(tracks_dict[k]) if k in tracks_dict else pyplt.get_BGR_color(
                        tracks_dict['default_tracks_color'])
                    len_tracks = tracks_dict['len_tracks']
                    r = tracks_dict['radius']
                    th = tracks_dict['thickness']

                    tracks_list = tracker_obj['tracks']
                    len_tracks = min(len_tracks, len(tracks_list))  # cant take more than the existing tracks
                    tracks_to_draw = tracks_list[-len_tracks:]  # take last 'len_tracks' tracks
                    for track in tracks_to_draw:
                        cv2.circle(cv_img, center=track, radius=r, color=tracks_color_bgr, thickness=th)
        if header_tl is not None:
            h = header_tl
            cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.TOP_LEFT.value, x_offset=h['x_offset'],
                           text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                           bg_font_scale=h['bg_font_scale'])
        if header_bl is not None:
            h = header_bl
            cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.BOTTOM_LEFT.value, x_offset=h['x_offset'],
                           text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                           bg_font_scale=h['bg_font_scale'])
        if header_tr is not None:
            h = header_tr
            cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.TOP_RIGHT.value, x_offset=h['x_offset'],
                           text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                           bg_font_scale=h['bg_font_scale'])
        if header_br is not None:
            h = header_br
            cvt.add_header(cv_img, header=h['text'], loc=pyplt.Location.BOTTOM_RIGHT.value, x_offset=h['x_offset'],
                           text_color=h['text_color'], with_rect=h['with_rect'], bg_color=h['bg_color'],
                           bg_font_scale=h['bg_font_scale'])
        return
